#########################################################################
# Copyright (C) 2016-2017  Shuaib Osman (sosman@investec.co.za)
# This file is part of RiskFlow.
#
# RiskFlow is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# RiskFlow is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RiskFlow.  If not, see <http://www.gnu.org/licenses/>.
#########################################################################

# import standard libraries

import os
import time
import logging
import itertools
import pandas as pd
import numpy as np
import tensorflow as tf

# needed to list available devices
from tensorflow.python.client import device_lib
# load up some useful data types
from collections import OrderedDict, namedtuple, defaultdict
# import the risk factors (also known as price factors)
from riskfactors import construct_factor
# import the stochastic processes
from stochasticprocess import construct_process
# import the currency/curve lookup factors 
from instruments import get_fxrate_factor, get_recovery_rate, get_interest_factor, get_survival_factor
# import the hessian function
from pricing import SensitivitiesEstimator
# import the documentation module
import documentation
# import the utils
import utils


class InstrumentExpired(Exception):
    def __init__(self, message):
        self.message = message


class Aggregation(object):
    def __init__(self, name):
        self.field = {'Reference': name}


class DealStructure(object):
    def __init__(self, obj, deal_level_mtm=False):
        # parent object - note that all parent objects MUST have an aggregate method that can aggregate partitions
        self.obj = utils.DealDataType(Instrument=obj, Factor_dep=None, Time_dep=None, Calc_res=None)
        # gather a list of deal dependencies
        self.dependencies = []
        # maintain a list of container objects
        self.sub_structures = []
        # do we need to create sub_partitions?
        self.sub_partitions = OrderedDict()
        # Do we want to store each deal level MTM explicitly?
        self.deal_level_mtm = deal_level_mtm

    def calc_time_dependency(self, base_date, deal, time_grid):
        # calculate the additional (dynamic) dates that this instrument needs to be evaluated at
        deal_time_dep = None
        try:
            reval_dates = deal.get_reval_dates(clip_expiry=True)
            if len(time_grid.scenario_dates) == 1:
                if len(reval_dates) > 0 and max(reval_dates) < base_date:
                    raise InstrumentExpired(deal.field.get('Reference', 'Unknown Instrument Reference'))
                deal_time_dep = time_grid.calc_deal_grid({base_date})
            else:
                deal_time_dep = time_grid.calc_deal_grid(reval_dates)
        except InstrumentExpired as e:
            logging.warning('skipping expired deal {0}'.format(e.args))

        return deal_time_dep

    def add_deal_to_structure(self, base_date, deal, static_offsets, stochastic_offsets, all_factors,
                              all_tenors, time_grid, calendars):
        """
        The logic is as follows: a structure contains deals - a set of deals are netted off and then the rules that the
        structure itself contains is applied.
        """
        deal_time_dep = self.calc_time_dependency(base_date, deal, time_grid)
        if deal_time_dep is not None:
            # calculate dependencies based on field names
            try:
                self.dependencies.append(
                    utils.DealDataType(Instrument=deal,
                                       Factor_dep=deal.calc_dependencies(base_date, static_offsets,
                                                                         stochastic_offsets,
                                                                         all_factors, all_tenors,
                                                                         time_grid, calendars),
                                       Time_dep=deal_time_dep,
                                       Calc_res=OrderedDict() if self.deal_level_mtm else None))
            except Exception as e:
                logging.error('{0}.{1} {2} - Skipped'.format(deal.field['Object'],
                                                             deal.field.get('Reference', '?'), e.args))

    def add_structure_to_structure(self, struct, base_date, static_offsets, stochastic_offsets,
                                   all_factors, all_tenors, time_grid, calendars):
        # get the dependencies
        struct_time_dep = self.calc_time_dependency(base_date, struct.obj.Instrument, time_grid)
        try:
            if struct_time_dep is not None:
                struct.obj = utils.DealDataType(
                    Instrument=struct.obj.Instrument,
                    Factor_dep=struct.obj.Instrument.calc_dependencies(base_date,
                                                                       static_offsets,
                                                                       stochastic_offsets,
                                                                       all_factors, all_tenors,
                                                                       time_grid, calendars),
                    Time_dep=struct_time_dep,
                    Calc_res=OrderedDict() if self.deal_level_mtm else None)
            # Structure object representing a netted set of cashflows
            self.sub_structures.append(struct)
        except Exception as e:
            logging.error('{0}.{1} {2} - Skipped'.format(struct.obj.Instrument.field['Object'],
                                                         struct.obj.Instrument.field.get('Reference', '?'), e.args))

    def build_partitions(self):
        # sort the data by the tags
        self.dependencies.sort(key=lambda x: x.Instrument.field['Tags'])
        # get the distinct tags
        unique_tags = set([x.Instrument.field['Tags'] for x in self.dependencies])
        if len(unique_tags) > 1:
            for tag in sorted(unique_tags):
                # TODO
                pass

    def resolve_structure(self, shared, time_grid):
        """
        Resolves the Structure
        """

        accum = None

        if self.sub_structures:
            # process sub structures
            for structure in self.sub_structures:
                struct = structure.resolve_structure(shared, time_grid)
                accum = struct if accum is None else accum + struct

        if self.dependencies and self.obj.Instrument.accum_dependencies:
            # accumulate the mtm's
            deal_tensors = 0.0

            for deal_data in self.dependencies:
                deal_ref = utils.Factor('Deal', (deal_data.Instrument.field.get('Reference', 'Unknown'),))

                with tf.name_scope(utils.check_scope_name(deal_ref)):
                    mtm = deal_data.Instrument.calculate(shared, time_grid, deal_data)
                    deal_tensors += mtm

            netting = deal_tensors
            accum = netting if accum is None else accum + netting

        # postprocessing code for working out the mtm of all deals, collateralization etc..
        if hasattr(self.obj.Instrument, 'post_process'):
            # the actual answer for this netting set
            accum = self.obj.Instrument.post_process(accum, shared, time_grid, self.obj, self.dependencies)

        return accum


class ScenarioTimeGrid(object):
    def __init__(self, cutoff_date, global_time_grid, base_date):
        scen_grid = global_time_grid.scen_time_grid
        offset = scen_grid.searchsorted((cutoff_date - base_date).days) + 1
        self.scen_time_grid = scen_grid[:offset]
        self.time_grid_years = self.scen_time_grid / utils.DAYS_IN_YEAR
        self.scenario_grid = global_time_grid.scenario_grid[:offset]


class TimeGrid(object):
    def __init__(self, scenario_dates, MTM_dates, base_MTM_dates):
        self.scenario_dates = scenario_dates
        self.base_MTM_dates = base_MTM_dates
        self.CurrencyMap = {}
        self.set_mtm_dates(MTM_dates)

    def set_mtm_dates(self, MTM_dates):
        self.mtm_dates = MTM_dates
        self.date_lookup = dict([(x, i) for i, x in enumerate(sorted(MTM_dates))])

    def calc_time_grid(self, time_in_days):
        dvt = np.concatenate(([1], np.diff(self.scen_time_grid), [1]))
        scen_index = self.scen_time_grid.searchsorted(time_in_days, side='right')
        index = (scen_index - 1).clip(0, self.scen_time_grid.size - 1)
        alpha = ((time_in_days - self.scen_time_grid[index]) / dvt[scen_index]).clip(0, 1)
        return np.dstack([alpha, time_in_days, index])[0]

    def set_base_date(self, base_date, delta=None):
        # leave the grids in terms of the number of days - note that it's possible to have the scenario_dates
        # the same as the mtm_dates (for more accurate margin period of risk on collateralized netting sets)
        self.mtm_time_grid = np.array([(x - base_date).days for x in sorted(self.mtm_dates)])
        self.scen_time_grid = np.array([(x - base_date).days for x in sorted(self.scenario_dates)])

        self.base_time_grid = set([self.date_lookup[x] for x in self.base_MTM_dates])
        self.time_grid = self.calc_time_grid(self.mtm_time_grid)

        # store the scenario time_grid
        self.scenario_grid = np.zeros((self.scen_time_grid.size, 3))
        self.scenario_grid[:, utils.TIME_GRID_MTM] = self.scen_time_grid
        self.scenario_grid[:, utils.TIME_GRID_ScenarioPriorIndex] = np.arange(self.scen_time_grid.size)

        # deal with the case that we need a very fine time_grid - note we do this after calculating the
        # scenario_grid as setting a non-null delta is a way to generate scenarios without calculating the
        # whole risk factor
        if delta is not None:
            delta_days, delta_tenors = delta
            delta_grid = np.union1d(np.arange(0, self.scen_time_grid.max(), delta_days), delta_tenors)
            self.scen_time_grid = np.union1d(self.scen_time_grid, delta_grid)

        self.time_grid_years = self.scen_time_grid / utils.DAYS_IN_YEAR

    def get_scenario_offset(self, days_from_base):
        prev_scen_index = self.scen_time_grid[self.scen_time_grid <= days_from_base].size - 1
        scenario_grid_delta = np.float64(
            (self.scen_time_grid[prev_scen_index + 1] - self.scen_time_grid[prev_scen_index]) if (
                    self.scen_time_grid.size > 1 and self.scen_time_grid.size > prev_scen_index + 1) else 1.0)
        return (days_from_base - self.scen_time_grid[prev_scen_index]) / scenario_grid_delta, prev_scen_index

    def set_currency_settlement(self, currencies):
        self.CurrencyMap = {}
        for currency, dates in currencies.items():
            settlement_dates = sorted([self.date_lookup[x] for x in dates if x in self.date_lookup])
            if settlement_dates:
                currency_lookup = np.zeros(self.mtm_time_grid.size, dtype=np.int32) - 1
                currency_lookup[settlement_dates] = np.arange(len(settlement_dates))
                self.CurrencyMap.setdefault(currency, currency_lookup)

    def calc_deal_grid(self, dates):
        try:
            dynamic_dates = self.base_time_grid.union([self.date_lookup[x] for x in dates])
        except KeyError as e:
            # if there is at least one reset date in the set of dates, then return it, else the deal has expired
            r = [self.date_lookup[x] for x in dates if x in self.date_lookup]
            if r:
                dynamic_dates = self.base_time_grid.union(r)
            else:
                if max(dates) < min(self.date_lookup.keys()):
                    raise InstrumentExpired(e)

                # include this instrument but don't bother pricing it through time
                return utils.DealTimeDependencies(self.mtm_time_grid, np.array([0]))

        # now construct the full deal grid 
        deal_time_grid = np.array(sorted(dynamic_dates))
        # find the last dynamic date - should be the expiry date
        expiry = self.date_lookup[max(dates)]
        # calculate the interpolation points etc.
        return utils.DealTimeDependencies(self.mtm_time_grid, deal_time_grid[deal_time_grid <= expiry])


class Calculation(object):

    def __init__(self, config, prec=np.float32):
        """
        Construct a new calculation - all calculations must setup the GPU ( by setting the precision )
        and by specifying a configuration (market data + porfolio)
        """

        self.config = config
        self.prec = prec
        self.time_grid = None
        self.shared_mem = None

        # the risk factor data
        self.static_factors = OrderedDict()
        self.static_ofs = OrderedDict()
        self.stoch_factors = OrderedDict()
        self.stoch_ofs = OrderedDict()
        self.all_factors = {}
        self.all_tenors = {}

        # these are used for implied calibration        
        self.implied_ofs = OrderedDict()

        self.dependent_factors = None
        self.base_date = None

        self.tenor_size = None
        self.tenor_offset = None

        # the deal structure
        self.netting_sets = None

        # generate slides from calculation?
        self.documentation = documentation.NoElaboration()
        # what's the name of this calculation?
        self.name = 'Unnamed'
        # performance and admin feedback
        self.calc_stats = {}
        # the calculation parameters (defined by calling execute)
        self.params = {}
        # Index for the gradients of this calculation (if requested)
        self.gradient_index = None

    def feed(self, **args):
        pass

    def execute(self, params):
        pass

    def make_factor_index(self, tensors):
        # need to match the indices back
        tenors = utils.get_tenors(self.all_factors)
        self.gradient_index = {}
        for var in tensors:
            name = var.name[:-2]
            factor_name = name.split('/')[-1]
            name_size, pad_size = tenors[factor_name].shape
            padding = 3 - pad_size
            indices = np.pad(tenors[factor_name], [[0, 0], [0, padding]], 'constant')
            self.gradient_index[name] = (indices, pad_size)

    def gradients_as_df(self, grad, header='Gradient', display_val=False):
        if isinstance(grad, dict):
            # get the factor values from all_factors if necessary
            factor_values = {utils.check_scope_name(k): v.factor if hasattr(v, 'factor') else v
                             for k, v in self.all_factors.items()} if display_val else {}
            hessian_index = ([], [])
            factor, rate, tenor, values = [], [], [], []
            for k, v in grad.items():
                # first derivative
                name = k.name[:-2]
                non_zero = np.where(v)[0]
                grad_index, index_len = self.gradient_index[name]
                hessian_index[0].append([name] * v.shape[0])
                hessian_index[1].append(grad_index)
                values.append(v[non_zero])
                rate.append([name] * non_zero.size)
                tenor.append(grad_index[non_zero])
                # store the actual factor value if required
                if display_val:
                    rate_name = name.split('/')[-1]
                    if rate_name in factor_values:
                        rate_non_zero = grad_index[non_zero][:, :index_len].astype(np.float64)
                        factor.append(factor_values[rate_name].current_value(rate_non_zero).flatten())
                    else:
                        sub_rate, param = rate_name.rsplit('.', 1)
                        factor.append(factor_values[sub_rate].current_value()[param].flatten())

            tenors = np.vstack(tenor)
            self.hessian_index = (np.hstack(hessian_index[0]), np.vstack(hessian_index[1]))

            data = {'Rate': np.hstack(rate), 'Tenor': tenors[:, 0], 'Tenor2': tenors[:, 1],
                    'Tenor3': tenors[:, 2], header: np.hstack(values)}
            index = ['Rate', 'Tenor', 'Tenor2', 'Tenor3']

            if display_val:
                data['Value'] = np.hstack(factor)

            df = pd.DataFrame(data).set_index(index).sort_index(level=[0, 1, 2, 3])
        else:
            # second derivative
            multi_index = []
            non_zero = np.where(grad)
            headers = [None, header]
            for extra_index, full_index in zip(headers, non_zero):
                index = np.unique(full_index)
                rate = self.hessian_index[0][index]
                tenor = self.hessian_index[1][index]
                m_index = [rate, tenor[:, 0], tenor[:, 1], tenor[:, 2]]
                if extra_index is not None:
                    m_index = [[extra_index] * rate.size] + m_index
                multi_index.append(pd.MultiIndex.from_arrays(m_index))

            values = grad[grad.any(axis=1)][:, grad.any(axis=0)]
            df = pd.DataFrame(values, index=multi_index[0],
                              columns=multi_index[1]).sort_index(level=[0, 1, 2, 3], axis=0)
        return df

    def set_deal_structures(self, deals, output, num_scenarios, batch_size, tagging=None, deal_level_mtm=False):
        partitioning = False
        for node in deals:
            # get the instrument
            instrument = node['instrument']
            # should we skip it?
            if node.get('Ignore') == 'True':
                continue
            # apply a tag if provided (used for partitioning)
            if tagging:
                instrument.field['Tags'] = tagging(instrument.field)
                partitioning = True
            if node.get('Children'):
                struct = DealStructure(instrument, deal_level_mtm=deal_level_mtm)
                logging.info('Analysing Group {0}'.format(instrument.field.get('Reference', '<undefined>')))
                self.set_deal_structures(node['Children'], struct, num_scenarios, batch_size, tagging, deal_level_mtm)
                output.add_structure_to_structure(struct, self.base_date, self.static_ofs, self.stoch_ofs,
                                                  self.all_factors, self.all_tenors, self.time_grid,
                                                  self.config.holidays)
                continue
            output.add_deal_to_structure(self.base_date, instrument, self.static_ofs, self.stoch_ofs, self.all_factors,
                                         self.all_tenors, self.time_grid, self.config.holidays)

        # check if we need to partition this structure
        if partitioning:
            output.build_partitions()


class Credit_Monte_Carlo(Calculation):
    documentation = ('Calculations', [
        'A profile is a curve $V(t)$ with values specified at a discrete set of future dates $0=t_0<t_1<...<t_m$ with',
        'values at other dates  obtained via linear interpolation or zero extrapolation i.e. if $t_{i-1}<t<t_i$ then',
        '$V(t)$ is a linear interpolation of $V(t_{i-1})$ and $V(t_i)$; otherwise $V(t)=0$.',
        '',
        'The valuation models described earlier are used to construct the profile. The profile dates $t_1,...,t_m$ are',
        'obtained by taking the following union:',
        '',
        '- The deal\'s maturity date.',
        '- The dates in the **Base Time Grid** up the the deal\'s maturity date.',
        '- Deal specific dates such as payment and exercise dates.',
        '',
        'Deal specific dates improve the accuracy of the profile by showing the effect of cashflows, exercises etc.',
        '',
        '### Aggregation',
        '',
        'If $U$ and $V$ are profiles, then the set $U+V$ is the union of profile dates $U$ and $V$. If $E$ is the',
        'credit exposure profile in reporting',
        'currency (**Currency**), then:',
        '',
        '$$E = \\sum_{d} V_d$$',
        '',
        'where $V_d$ is the valuation profile of the $d^{th}$ deal. Note that Netting is always assumed to be',
        '**True**.',
        '',
        '#### Peak Exposure',
        '',
        'This is the simulated exposure at percentile $q$ where $0<q<1$ (typically q=.95 or .99).',
        '',
        '#### Expected Exposure',
        '',
        'This is the profile defined by taking the average of the positive simulated exposures i.e. for each profile',
        'date $t$,',
        '',
        '$$\\bar E(t)=\\frac{1}{N}\\sum_{k=1}^N \\max(E(t)(k),0).$$'
        '',
        '#### Exposure Deflation',
        '',
        'Exposure at time $t$ is simulated in units of the time $t$ reporting currency. Exposure deflation converts',
        'this to time $0$ reporting currency i.e.',
        '',
        '$$V^*(t)=\\frac{V(t)}{\\beta(t)}$$',
        '',
        'where',
        '',
        '$$\\beta(t)=\\exp\\Big(\\int_0^t r(s)ds\\Big).$$',
        '',
        'This can be approximated by:',
        '',
        '$$\\beta(t)=\\prod_{i=0}^n\\frac{1}{D(s_{i-1},s_i)}$$',
        '',
        'where $0=s_0<...<s_n=t$. The discrete set of dates $s_1,...,s_{n-1}$ are model-dependent.'
        '',
        '### Credit Valuation Adjustment',
        '',
        'This represents the adjustment to the market value of the portfolio accounting for the risk of default. Only',
        'unilateral CVA (i.e accounting for the counterparty risk of default but ignoring the potential default of',
        'the investor) is calculated. It is given by:',
        '',
        '$$C=\\Bbb E(L(\\tau)),$$',
        '',
        'where the expectation is taken with respect to the risk-neutral measure, and ',
        '',
        '$$L(t)=(1-R)\\max(E^*(t),0),$$',
        '',
        'with:',
        '',
        '- $R$ the counterparty recovery rate',
        '- $\\tau$ the counterparty time to default',
        '- $E^*(t)$ the exposure at time $t$ deflated by the money market account.',
        '',
        'If **Deflate Stochastically** is **No** then the deflated expected exposure is assumed to be deterministic',
        'i.e. $E^*(t)=E(t)D(0,t)$. Note that if $T$ is the end date of the portfolio exposure then $E^*(t)=0$ for',
        '$t>T$.',
        '',
        'Now,',
        '',
        '$$\\Bbb E(L(\\tau))=\\Bbb E\\Big(\\int_0^T L(t)(-dH(t))\\Big),$$',
        '$$H(t)=\\exp\\Big(-\\int_0^t h(u)du\\Big)$$',
        '',
        'where $h(t)$ is the stochastic hazard rate. There are two ways to calculate the expectation:',
        '',
        'If **Stochastic Hazard** is **No** then $H(t)=\\Bbb P(\\tau > t)=S(0,t)$, the risk neutral survival',
        'probability to time $t$ and',
        '',
        '$$C=\\int_0^T \\Bbb E(L(t))(-dH(t))\\approx\\sum_{i=1}^m C_i,$$',
        '',
        'with',
        '',
        '$$C_i=\\Big(\\frac{\\Bbb E(L(t_{i-1}))+\\Bbb E(L(t_i))}{2}\\Big)(S(0,t_{i-1})-S(0,t_i))$$',
        '',
        'and $0=t_0<...<t_m=T$ are the time points on the exposure profile. Note that the factor models used should',
        'be risk-neutral to give risk neutral simulations of $\\Bbb E^*(t)$.',
        '',
        'If **Stochastic Hazard** is **Yes** then $S(t,u)$ is the simulated survival probability at time $t$ for',
        'maturity $u$ and is related to $H$ by',
        '',
        '$$S(t,u)=\\Bbb E(\\frac{H(u)}{H(t)}\\vert\\mathcal F(t)).$$',
        '',
        'where $\\mathcal F$ is the filtration given by the risk factor processes. For small $u-t$, the approximation',
        '$H(u)\\approx H(t)S(t,u)$ is accurate so that',
        '',
        '$$C\\approx\\sum_{i=1}^m C_i$$',
        '',
        'and',
        '',
        '$$C_i=\\Bbb E\\Big[\\Big(\\frac{L(t_{i-1}))+L(t_i)}{2}\\Big)(H_{i-1}-H_i)\\Big],$$',
        '',
        'again, $0=t_0<...<t_m=T$ are the time points on the exposure profile and',
        '',
        '$$H_i=S(0,t_1)S(t_1,t_2)...S(t_{i-1},t_i).$$',
        '',
        '### Funding Valuation Adjustment',
        '',
        'Posting (or recieving) collateral can imply a funding cost (or benefit) when there is a spread between a',
        'party\'s interal cost of funding and the contractual interest rate paid on the collateral balance. The',
        'discounted expectation of this cost (or benefit) summed across all time horizons and scenarios constitutes',
        'a funding value adjustment and can be expressed as:',
        '',
        '$$\\frac{1}{m}\\sum_{j=1}^m \\sum_{k=0}^{T-1} B_j(t_k)S_j(t_k)\\Big(\\frac{D_j^c(t_k,t_{k+1})}'
        '{D_j^f(t_k,t_{k+1})}-1\\Big)D_j^c(0,t_k),$$',
        '',
        'where',
        '',
        '- $B_j(t)$ is the number of units of the collateral portfolio for scenario $j$ at time $t$',
        '- $S_j(t)$ is the base currecy value of one unit of the collateral asset for scenario $j$ at time $t$',
        '- $D_j^c(t)$ is the discount rate for the collateral rate at time t for scenario $j$',
        '- $D_j^f(t)$ is the discount rate for the funding rate at time t for scenario $j$',
        '',
        'Note that only cash collateral is supported presently although this can be extended.'
    ])

    def __init__(self, config, **kwargs):
        super(Credit_Monte_Carlo, self).__init__(config, **kwargs)
        self.cholesky = None
        self.stochastic_factors = None
        self.reset_dates = None
        self.settlement_currencies = None

        # global variables to store the state of the graph between calculations
        self.shared_mem_class = namedtuple('shared_mem', 't_Buffer t_random_numbers \
                                                        t_Scenario_Buffer t_Static_Buffer t_Cashflows \
                                                        t_Feed_dict t_Credit gpus riskneutral precision \
                                                        simulation_batch Report_Currency')

        # used to store the Scenarios (if requested)
        self.Scenarios = {}
        # used to store any jacobian matrices
        self.jacobians = {}
        # used to store the resulting tensors of the calculation
        self.tensors = OrderedDict()
        # implied factors
        self.implied_factors = OrderedDict()
        # we represent the calc as a combination of static, stochastic and implied parameters
        self.stoch_values = []
        self.stoch_var = []
        self.static_values = []
        self.static_var = []
        self.implied_var = []
        self.implied_values = []
        # potentially store the full list of variables
        self.all_var = None

        self.stoch_ofs = OrderedDict()
        self.static_ofs = OrderedDict()
        self.implied_ofs = OrderedDict()

    def update_factors(self, params, base_date):
        dependent_factors, stochastic_factors, implied_factors, reset_dates, settlement_currencies = \
            self.config.calculate_dependencies(params, base_date, self.input_time_grid)

        # update the time grid
        logging.info('Updating timegrid')
        self.update_time_grid(base_date, reset_dates, settlement_currencies,
                              dynamic_scenario_dates=params.get('Dynamic_Scenario_Dates', 'No') == 'Yes')

        # now construct the stochastic factors and static factors for the simulation
        self.stoch_factors.clear()

        for price_model, price_factor in stochastic_factors.items():
            factor_obj = construct_factor(price_factor, self.config.params['Price Factors'])
            implied_factor = implied_factors.get(price_model)
            try:
                if implied_factor:
                    implied_obj = construct_factor(implied_factor, self.config.params['Price Factors'])
                    self.implied_factors[implied_factor] = implied_obj
                else:
                    implied_obj = None
            except KeyError as e:
                logging.warning(
                    'Implied Factor {0} missing in market data file - Attempting to Proxy with EUR-MASTER'.format(
                        e.args))
                # Hardcoding - Fix this - TODO
                proxy_factor = utils.Factor(implied_factor.type, ('EUR-MASTER',))
                self.config.params['Price Factors'][utils.check_tuple_name(implied_factor)] = \
                    self.config.params['Price Factors'][utils.check_tuple_name(proxy_factor)]
                implied_obj = construct_factor(implied_factor, self.config.params['Price Factors'])

            self.stoch_factors[price_factor] = construct_process(price_model.type,
                                                                 factor_obj,
                                                                 self.config.params['Price Models']
                                                                 [utils.check_tuple_name(price_model)],
                                                                 implied_obj)

        self.static_factors = OrderedDict()
        for price_factor in set(dependent_factors).difference(stochastic_factors.values()):
            try:
                self.static_factors.setdefault(
                    price_factor, construct_factor(price_factor, self.config.params['Price Factors']))
            except KeyError as e:
                logging.warning('Price Factor {0} missing in market data file - skipping'.format(e.args))

        self.all_factors = self.stoch_factors.copy()
        self.all_factors.update(self.static_factors)
        self.all_factors.update(self.implied_factors)
        self.num_factors = sum([v.num_factors() for v in self.stoch_factors.values()])

        # get the tenor offset (if any)
        tenor_offset = params.get('Tenor_Offset', 0.0)

        # now get the stochastic risk factors ready - these will be generated from the price models
        with tf.name_scope("Stochastic_Input"):
            for key, value in self.stoch_factors.items():
                if key.type not in utils.DimensionLessFactors:
                    # check if there are any implied factors linked here
                    with tf.name_scope("Implied/"):
                        if hasattr(value, 'implied'):
                            vals, vars = [], OrderedDict()
                            self.implied_ofs.setdefault(key, len(self.implied_var))
                            for param_name, param_value in value.implied.current_value().items():
                                factor_name = utils.Factor(value.implied.__class__.__name__,
                                                           key.name + (param_name,))
                                vals.append(param_value)
                                vars[param_name] = tf.placeholder(self.prec, shape=len(param_value),
                                                                  name=utils.check_scope_name(factor_name))
                            self.implied_values.append(vals)
                            self.implied_var.append(vars)

                    # record the offset of this risk factor
                    self.stoch_ofs.setdefault(key, len(self.stoch_var))
                    current_val = value.factor.current_value(offset=tenor_offset)
                    self.stoch_values.append(current_val)
                    self.stoch_var.append(
                        tf.placeholder(self.prec,
                                       shape=len(current_val),
                                       name=utils.check_scope_name(key)))

        # and then get the the static risk factors ready - these will just be looked up
        with tf.name_scope("Static_Input"):
            for key, value in self.static_factors.items():
                if key.type not in utils.DimensionLessFactors:
                    # record the offset of this risk factor
                    self.static_ofs.setdefault(key, len(self.static_var))
                    current_val = value.current_value(offset=tenor_offset)
                    self.static_values.append(current_val)
                    self.static_var.append(
                        tf.placeholder(self.prec,
                                       shape=len(current_val),
                                       name=utils.check_scope_name(key)))

        # update the correlations
        self.update_correlation()
        # check if we need gradients for any xVA
        greeks = np.any([params[x].get('Gradient', 'No') == 'Yes' for x in params.keys() if x.endswith('VA')])

        # setup the device and allocate memory
        self.__refresh_shared_mem(params['Random_Seed'],
                                  params.get('NoModel', 'Constant'),
                                  params['Currency'],
                                  calc_greeks=greeks)

        # calculate a reverse lookup for the tenors and store the daycount code
        self.all_tenors = utils.update_tenors(self.base_date, self.all_factors)

        # now initialize all stochastic factors
        for key, value in self.stoch_factors.items():
            if key.type not in utils.DimensionLessFactors:
                with tf.name_scope(utils.check_scope_name(key)):
                    implied_index = self.implied_ofs.get(key, -1)

                    value.precalculate(
                        base_date, ScenarioTimeGrid(
                            dependent_factors[key], self.time_grid, base_date),
                        self.stoch_var[self.stoch_ofs[key]],
                        self.shared_mem, self.process_ofs[key],
                        implied_tensor=self.implied_var[implied_index] if implied_index > -1 else None)

        # now check if any of the stochastic processes depend on other processes
        for key, value in self.stoch_factors.items():
            if key.type not in utils.DimensionLessFactors:
                # precalculate any values for the stochastic process
                value.calc_references(key, self.static_ofs, self.stoch_ofs, self.all_tenors, self.all_factors)

        # store the state info
        self.dependent_factors, self.stochastic_factors = dependent_factors, stochastic_factors

    def update_time_grid(self, base_date, reset_dates, settlement_currencies, dynamic_scenario_dates=False):
        # work out the scenario and dynamic dates
        dynamic_dates = set([x for x in reset_dates if x > base_date])
        base_mtm_dates = self.config.parse_grid(base_date, max(dynamic_dates), self.input_time_grid)
        mtm_dates = base_mtm_dates.union(dynamic_dates)
        if dynamic_scenario_dates:
            scenario_dates = mtm_dates
        else:
            scenario_dates = self.config.parse_grid(base_date, max(dynamic_dates), self.input_time_grid,
                                                    past_max_date=True)

        # setup the scenario and base time grids
        self.time_grid = TimeGrid(scenario_dates, mtm_dates, base_mtm_dates)
        self.base_date = base_date
        self.reset_dates = reset_dates
        self.time_grid.set_base_date(base_date)

        # Set the settlement dates
        self.time_grid.set_currency_settlement(settlement_currencies)
        self.settlement_currencies = settlement_currencies

    def update_correlation(self):
        # create the correlation matrix
        correlation_matrix = np.eye(self.num_factors, dtype=self.prec)

        # prepare the correlation matrix (and the offsets of each stochastic process)
        correlation_factors = []
        self.process_ofs = OrderedDict()
        for key, value in self.stoch_factors.items():
            proc_corr_type, proc_corr_factors = value.correlation_name
            for sub_factors in proc_corr_factors:
                # record the offset of this factor model
                self.process_ofs.setdefault(key, len(correlation_factors))
                # record the name of needed correlation lookup
                correlation_factors.append(utils.Factor(proc_corr_type, key.name + sub_factors))

        for index1 in range(self.num_factors):
            for index2 in range(index1 + 1, self.num_factors):
                factor1, factor2 = utils.check_tuple_name(correlation_factors[index1]), utils.check_tuple_name(
                    correlation_factors[index2])
                key = (factor1, factor2) if (factor1, factor2) in self.config.params['Correlations'] else (
                    factor2, factor1)
                rho = self.config.params['Correlations'].get(key, 0.0) if factor1 != factor2 else 1.0
                correlation_matrix[index1, index2] = rho
                correlation_matrix[index2, index1] = rho

        # need to do cholesky
        eigval, eigvec = np.linalg.eig(correlation_matrix.astype(np.float64))
        if not (eigval > 1e-8).all():
            # matrix not positive definite - find a close positive definite matrix
            if self.config.params['System Parameters']['Correlations_Healing_Method'] == 'Eigenvalue_Raising':
                logging.warning('Correlation matrix (size {0}) not positive definite - raising eigenvalues'.format(
                    correlation_matrix.shape))
                P_plus_B = eigvec.dot(np.diag(np.maximum(eigval, 1e-4))).dot(eigvec.T)
                diagonal_norm = np.diag(1.0 / np.sqrt(P_plus_B.diagonal()))
                new_correlation_matrix = diagonal_norm.dot(P_plus_B).dot(diagonal_norm)
            else:
                logging.warning('Correlation matrix (size {0}) not positive definite - alternating Projections'.format(
                    correlation_matrix.shape))

                C = correlation_matrix.astype(np.float64).copy()
                B = correlation_matrix.astype(np.float64).copy()

                # don't do more than 100 iterations - if we need to do this much, behaviour is undefined
                for k in range(100):
                    eigval, eigvec = np.linalg.eig(B)
                    P_plus_B = eigvec.dot(np.diag(np.maximum(eigval, 1e-4))).dot(eigvec.T)
                    nC = P_plus_B + np.diag(1.0 - P_plus_B.diagonal())
                    D = nC - P_plus_B
                    B += D

                    # exit early
                    if np.abs(C - nC).max() < 1e-08 * np.abs(nC).max():
                        break

                    C = nC

                new_correlation_matrix = nC

            correlation_matrix = new_correlation_matrix

        self.correlation_matrix = correlation_matrix
        with tf.name_scope("Correlation_Input"):
            self.input_correlation = tf.placeholder(self.prec, shape=correlation_matrix.shape,
                                                    name='correlation_matrix')
            # set the cholesky decomp
            self.cholesky = tf.cholesky(self.input_correlation)

    def __refresh_shared_mem(self, seed, nomodel, reporting_currency, calc_greeks=False, debug_batch_size=128):
        # setup the Feed dictionary - the inputs needed to evaluate the graph
        feed = {self.input_correlation: self.correlation_matrix}
        feed.update(dict(zip(self.stoch_var, self.stoch_values)))
        feed.update(dict(zip(self.static_var, self.static_values)))
        feed.update(
            dict(zip(itertools.chain(*[x.values() for x in self.implied_var]),
                     itertools.chain(*self.implied_values))))
        if not isinstance(self.batch_size, int):
            feed.update({self.batch_size: debug_batch_size})
            # this makes a difference in tensorflow's internals for some reason
            # might be possible to remove this once fixed
            cashflow_size = [self.batch_size]
        else:
            cashflow_size = self.batch_size

        if calc_greeks:
            implied_vars = list(itertools.chain(*[x.values() for x in self.implied_var]))
            self.all_var = implied_vars #self.stoch_var + self.static_var + implied_vars
            # build our index                 
            self.make_factor_index(self.all_var)

        # Now apply that cholesky to our random numbers.
        # There may be a better way to do this as the cholesky is lower triangular

        with tf.name_scope('Setup'):
            self.shared_mem = self.shared_mem_class(
                t_Buffer={},
                t_random_numbers=tf.reshape(
                    tf.matmul(
                        self.cholesky,
                        tf.random_normal(
                            [self.num_factors, self.batch_size * self.time_grid.scen_time_grid.size],
                            seed=seed, dtype=self.prec),
                    ),
                    (self.num_factors, self.time_grid.scen_time_grid.size, -1), name='d_random_numbers'),
                t_Scenario_Buffer=[None] * len(self.stoch_factors),
                t_Static_Buffer=self.static_var,
                t_Cashflows={k: {t_i: tf.zeros(cashflow_size, dtype=self.prec) for t_i in np.where(v >= 0)[0]}
                             for k, v in self.time_grid.CurrencyMap.items()},
                t_Feed_dict=feed,
                t_Credit={},
                gpus=[x.name for x in device_lib.list_local_devices() if x.device_type == 'GPU'],
                riskneutral=nomodel == 'RiskNeutral',
                precision=self.prec,
                simulation_batch=self.batch_size,
                Report_Currency=get_fxrate_factor(utils.check_rate_name(reporting_currency),
                                                  self.static_ofs, self.stoch_ofs)
            )

    def feed(self, batch_size, correlation=None, stoch_vars={}, static_vars={},
             options=None, run_meta=None, writer=None):

        # record the performance stats
        self.calc_stats['Tensor_Execution_Time'] = time.clock()

        # config = tf.ConfigProto(device_count = {'GPU': 0})
        config = tf.ConfigProto(allow_soft_placement=True)

        # clear the output
        self.output = {}
        # update the feed
        feed = self.shared_mem.t_Feed_dict.copy()
        # copy the static params
        feed.update({self.graph.get_tensor_by_name('Static_Input/{0}:0'.format(k)): v for
                     k, v in static_vars.items()})
        # copy the stoch params
        feed.update({self.graph.get_tensor_by_name('Stochastic_Input/{0}:0'.format(k)): v for
                     k, v in stoch_vars.items()})
        # update the correlation if necessary
        if correlation is not None:
            # note - I don't check if the new matrix is positive definite - this will break if it is
            feed.update({self.input_correlation: correlation})

        # change the batch size
        if not isinstance(self.batch_size, int):
            feed[self.batch_size] = batch_size

        with tf.Session(graph=self.graph, config=config) as sess:
            output = defaultdict(list)
            tensors = list(self.tensors.values())
            for run in range(self.params['Simulation_Batches']):
                # get the output tensors
                for k, v in zip(self.tensors.keys(), sess.run(tensors, feed_dict=feed,
                                                              options=options, run_metadata=run_meta)):
                    output[k].append(v)

                # fetch cashflows if necessary
                if self.params.get('Generate_Cashflows', 'No') == 'Yes':
                    dates = np.array(sorted(self.time_grid.mtm_dates))
                    for currency, values in self.shared_mem.t_Cashflows.items():
                        curr = sess.run(values, feed_dict=feed, options=options, run_metadata=run_meta)
                        cash_index = dates[sorted(curr.keys())]
                        output.setdefault('cashflows', {}).setdefault(currency, []).append(
                            pd.DataFrame([v for _, v in sorted(curr.items())], index=cash_index))

                # add any scenarios if necessary
                if self.params.get('Calc_Scenarios', 'No') == 'Yes':
                    for key, value in self.stoch_factors.items():
                        output.setdefault('scenarios', {}).setdefault(key, []).append(
                            sess.run(self.shared_mem.t_Scenario_Buffer[self.stoch_ofs[key]],
                                     feed_dict=feed, options=options, run_metadata=run_meta))
                if writer:
                    writer.add_run_metadata(run_meta, 'run {}'.format(run), run)

        for result, data in output.items():
            if result == 'scenarios':
                self.output.setdefault('scenarios', {k: np.concatenate(v, axis=-1) for k, v in data.items()})
            elif result == 'cashflows':
                self.output.setdefault('cashflows', {k: pd.concat(v, axis=1) for k, v in data.items()})
            elif result in ['cva', 'collva', 'fva']:
                self.output.setdefault(result, np.array(data, dtype=np.float64).mean())
            elif result == 'collva_t':
                self.output.setdefault(result, np.array(data, dtype=np.float64).mean(axis=0))
            elif result in ['grad_cva', 'grad_collva', 'grad_fva']:
                grad = dict.fromkeys(data[0].keys())
                for k in grad.keys():
                    values = np.array([v[k].astype(np.float64) for v in data])
                    grad[k] = np.mean(values, axis=0)
                self.output.setdefault(result, grad)
            elif result in ['grad_cva_hessian']:
                self.output.setdefault(result, np.stack(data).astype(np.float64).mean(axis=0))
            else:
                self.output.setdefault(result, np.concatenate(data, axis=-1).astype(np.float64))

        self.calc_stats['Tensor_Execution_Time'] = time.clock() - self.calc_stats['Tensor_Execution_Time']
        return self.output

    def execute(self, params, feedgraph=True):
        # get the rundate
        base_date = pd.Timestamp(params['Run_Date'])
        # check if we need to produce a slideshow (i.e. documentation)
        if params['Generate_Slideshow'] == 'Yes':
            # we need to generate scenarios to draw pictures . . .
            params['Calc_Scenarios'] = 'Yes'
            # check if we need to compare the results to an externally produced result
            compare_output = pd.read_csv(params['PFE_Recon_File'], index_col=0, parse_dates=True) \
                if params['PFE_Recon_File'] else None
            # set the name of this calculation
            self.name = params['calc_name'][0]
            # create some slides
            self.documentation = documentation.Elaboration(self.name)
        else:
            compare_output = ''
            self.documentation = documentation.NoElaboration()

        # Define the base and scenario grids
        self.input_time_grid = params['Time_grid']
        self.numscenarios = params['Batch_Size'] * params['Simulation_Batches']

        # reset the default graph
        tf.reset_default_graph()
        # create a TF graph to represent this calculation
        self.graph = tf.Graph()
        # set the precision in the utils module - Nasty hack - TODO
        utils.Default_Precision = self.prec
        # store the params                
        self.params = params
        # Dynamic Batch?
        dynamic = params.get('Dynamic', 'No') == 'Yes'
        # build the TF graph
        with self.graph.as_default() as g:
            self.batch_size = tf.placeholder(tf.int32, shape=[], name='Batch_Size') if dynamic else params['Batch_Size']
            self.calc_stats['Factor_Setup_Time'] = time.clock()
            self.update_factors(params, base_date)

            # record the setup time
            self.calc_stats['Factor_Setup_Time'] = time.clock() - self.calc_stats['Factor_Setup_Time']

            # now that the memory is setup, all python objects have had a chance to sync with the device
            # (in terms of indices, offsets etc.)
            self.calc_stats['Deal_Setup_Time'] = time.clock()
            self.netting_sets = DealStructure(Aggregation('root'))
            self.set_deal_structures(self.config.deals['Deals']['Children'], self.netting_sets, self.numscenarios,
                                     params['Batch_Size'], tagging=None,
                                     deal_level_mtm=params.get('DealLevel', False))

            # record the (pure python) dependency setup time
            self.calc_stats['Deal_Setup_Time'] = time.clock() - self.calc_stats['Deal_Setup_Time']

            # record how long it took to build the graph
            self.calc_stats['Graph_Setup_Time'] = time.clock()

            # simulate the price factors
            for key, value in self.stoch_factors.items():
                with tf.name_scope(utils.check_scope_name(key)):
                    self.shared_mem.t_Scenario_Buffer[self.stoch_ofs[key]] = value.generate(self.shared_mem)

            # construct the valuations
            self.tensors['mtm'] = self.netting_sets.resolve_structure(self.shared_mem, self.time_grid)

            # now calculate all the valuation adjustments (if necessary)
            if 'CollVA' in params and 'Funding' in self.shared_mem.t_Credit:
                with tf.name_scope('CollVA'):
                    self.tensors['collva_t'] = tf.reduce_mean(self.shared_mem.t_Credit['Funding'], axis=1)
                    self.tensors['collva'] = tf.reduce_sum(self.tensors['collva_t'])
                    self.tensors['collateral'] = self.shared_mem.t_Credit['Collateral']

                    if params['CollVA'].get('Gradient', 'No') == 'Yes':
                        # calculate all the derivatives of collva
                        self.tensors['grad_collva'] = SensitivitiesEstimator(
                            self.tensors['collva'], self.all_var).grad

            if 'FVA' in params:
                time_grid = self.netting_sets.sub_structures[0].obj.Time_dep.deal_time_grid
                mtm_grid = self.time_grid.mtm_time_grid[time_grid]

                funding = get_interest_factor(utils.check_rate_name(params['FVA']['Funding_Interest_Curve']),
                                              self.static_ofs, self.stoch_ofs, self.all_tenors)
                riskfree = get_interest_factor(utils.check_rate_name(params['FVA']['Risk_Free_Curve']),
                                               self.static_ofs, self.stoch_ofs, self.all_tenors)

                if 'Counterparty' in params['FVA']:
                    survival = get_survival_factor(utils.check_rate_name(params['FVA']['Counterparty']),
                                                   self.static_ofs,
                                                   self.stoch_ofs, self.all_tenors)
                    surv = utils.calc_time_grid_curve_rate(survival, np.zeros((1, 3)), self.shared_mem)
                    St_T = tf.squeeze(tf.exp(-surv.gather_weighted_curve(
                        self.shared_mem, mtm_grid.reshape(1, -1), multiply_by_time=False)), axis=0)
                else:
                    St_T = tf.constant([[1.0]], self.prec)

                with tf.name_scope('FVA'):
                    deflation = utils.calc_time_grid_curve_rate(riskfree, np.zeros((1, 3)), self.shared_mem)
                    DF_rf = tf.squeeze(tf.exp(-deflation.gather_weighted_curve(
                        self.shared_mem, mtm_grid.reshape(1, -1))), axis=0)

                    if params['FVA'].get('Stochastic_Funding', 'No') == 'Yes':
                        Vk_plus_ti = tf.nn.relu(self.tensors['mtm'] * DF_rf)
                        Vk_minus_ti = tf.nn.relu(-self.tensors['mtm'] * DF_rf)
                        Vk_star_ti_p = (Vk_plus_ti[1:] + Vk_plus_ti[:1]) / 2
                        Vk_star_ti_m = (Vk_minus_ti[1:] + Vk_minus_ti[:1]) / 2

                        delta_scen_t = np.hstack((0.0, np.diff(mtm_grid))).astype(self.prec).reshape(-1, 1)

                        discount_fund = utils.calc_time_grid_curve_rate(funding,
                                                                        self.time_grid.time_grid[time_grid],
                                                                        self.shared_mem)
                        discount_rf = utils.calc_time_grid_curve_rate(riskfree, self.time_grid.time_grid[time_grid],
                                                                      self.shared_mem)

                        delta_fund_rf = tf.squeeze(
                            tf.exp(discount_fund.gather_weighted_curve(self.shared_mem, delta_scen_t)) -
                            tf.exp(discount_rf.gather_weighted_curve(self.shared_mem, delta_scen_t)), axis=1) * St_T

                        FCA_t = tf.reduce_mean(delta_fund_rf[1:] * Vk_star_ti_p, axis=1)
                        FCA = tf.reduce_sum(FCA_t)
                        FBA_t = tf.reduce_mean(delta_fund_rf[1:] * Vk_star_ti_m, axis=1)
                        FBA = tf.reduce_sum(FBA_t)
                    else:
                        pass

                    self.tensors['fva'] = FCA - FBA

                    if params['FVA'].get('Gradient', 'No') == 'Yes':
                        # calculate all the derivatives of fva
                        self.tensors['grad_fva'] = SensitivitiesEstimator(
                            self.tensors['fva'], self.all_var).grad

            if 'CVA' in params:
                discount = get_interest_factor(utils.check_rate_name(params['Deflation_Interest_Rate']),
                                               self.static_ofs, self.stoch_ofs, self.all_tenors)
                survival = get_survival_factor(utils.check_rate_name(params['CVA']['Counterparty']),
                                               self.static_ofs,
                                               self.stoch_ofs, self.all_tenors)
                recovery = get_recovery_rate(utils.check_rate_name(params['CVA']['Counterparty']), self.all_factors)
                # only looks at the first netting set - should be fine . . .
                time_grid = self.netting_sets.sub_structures[0].obj.Time_dep.deal_time_grid

                # Calculates unilateral CVA with or without stochastic deflation.
                with tf.name_scope('CVA'):
                    mtm_grid = self.time_grid.mtm_time_grid[time_grid]
                    delta_scen_t = np.hstack((0.0, np.diff(mtm_grid))).astype(self.prec)

                    if params['CVA']['Deflate_Stochastically'] == 'Yes':
                        zero = utils.calc_time_grid_curve_rate(discount, self.time_grid.time_grid[time_grid],
                                                               self.shared_mem)
                        Dt_T = tf.exp(-tf.cumsum(tf.squeeze(zero.gather_weighted_curve(
                            self.shared_mem, delta_scen_t.reshape(-1, 1))), axis=0))
                    else:
                        zero = utils.calc_time_grid_curve_rate(discount, np.zeros((1, 3)), self.shared_mem)
                        Dt_T = tf.squeeze(tf.exp(-zero.gather_weighted_curve(
                            self.shared_mem, mtm_grid.reshape(1, -1))), axis=0)

                    pv_exposure = tf.nn.relu(self.tensors['mtm'] * Dt_T)

                    if params['CVA']['Stochastic_Hazard'] == 'Yes':
                        surv = utils.calc_time_grid_curve_rate(survival, self.time_grid.time_grid[time_grid],
                                                               self.shared_mem)
                        St_T = tf.exp(-tf.cumsum(tf.squeeze(surv.gather_weighted_curve(
                            self.shared_mem, delta_scen_t.reshape(-1, 1), multiply_by_time=False), axis=1),
                            axis=0))
                    else:
                        surv = utils.calc_time_grid_curve_rate(survival, np.zeros((1, 3)), self.shared_mem)
                        St_T = tf.squeeze(tf.exp(-surv.gather_weighted_curve(
                            self.shared_mem, mtm_grid.reshape(1, -1), multiply_by_time=False)), axis=0)

                    prob = St_T[:-1] - St_T[1:]

                    self.tensors['cva'] = (1.0 - recovery) * tf.reduce_sum(
                        tf.reduce_mean(0.5 * (pv_exposure[1:] + pv_exposure[:-1]) * prob, axis=1), axis=0)

                    if params['CVA'].get('Gradient', 'No') == 'Yes':
                        # potentially fetch ir jacobian matrices for base curves
                        base_ir_curves = [x for x in self.stoch_ofs.keys() if
                                          x.type == 'InterestRate' and len(x.name) == 1]
                        self.jacobians = {}
                        for ir_factor in base_ir_curves:
                            jacobian_factor = utils.Factor('InterestRateJacobian', ir_factor.name)
                            ir_curve = self.stoch_factors[utils.Factor('InterestRate', ir_factor.name)].factor
                            var_name = 'Stochastic_Input/{0}:0'.format(utils.check_tuple_name(ir_factor))
                            try:
                                jac = construct_factor(jacobian_factor, self.config.params['Price Factors'])
                                jac.update(ir_curve)
                                self.jacobians[var_name] = jac.current_value()
                                logging.info('jacobian present for {0} - will attempt inverse bootstrap'.format(
                                    utils.check_tuple_name(ir_factor)))
                            except KeyError as e:
                                pass

                        # calculate all the derivatives of cva
                        sensitivity = SensitivitiesEstimator(self.tensors['cva'], self.all_var)
                        self.tensors['grad_cva'] = sensitivity.grad
                        # store the size of the Gradient
                        self.calc_stats['Gradient_Vector_Size'] = sensitivity.P
                        if params['CVA'].get('Hessian', 'No') == 'Yes':
                            # calculate the hessian matrix - warning - this will probably kill your box
                            self.tensors['grad_cva_hessian'] = sensitivity.get_H_op(self.prec)

            self.calc_stats['Graph_Setup_Time'] = time.clock() - self.calc_stats['Graph_Setup_Time']

        results = {'Netting': self.netting_sets, 'Stats': self.calc_stats, 'Jacobians': self.jacobians}

        if params.get('Debug', 'No') != 'No' and os.path.isdir(params['Debug']):
            # record the number of tensors used (lower is better) - this is an expensive op
            self.calc_stats['Graph_Nodes'] = len(self.graph.as_graph_def().node)
            run_options = tf.RunOptions(trace_level=tf.RunOptions.FULL_TRACE)
            run_metadata = tf.RunMetadata()
            writer = tf.summary.FileWriter(params['Debug'], graph=self.graph)
            results['Results'] = self.feed(params['Batch_Size'], options=run_options, run_meta=run_metadata,
                                           writer=writer)
        else:
            results['Results'] = self.feed(params['Batch_Size']) if feedgraph else None

        return results


class Base_Revaluation(Calculation):
    """Simple deal revaluation - Use this to reconcile with the source system"""
    documentation = ('Calculations',
                     ['This applies the valuation models mentioned earlier to the portfolio per deal.',
                      '',
                      'The only inputs are:',
                      '',
                      '- **Currency** of the output.',
                      '- **Run_Date** at which the marketdata should be applied (i.e. $t_0$)',
                      '',
                      'The output is a dictionary containing the DealStructure and the calculation computation',
                      'statistics.'
                      ])

    def __init__(self, config, **kwargs):
        super(Base_Revaluation, self).__init__(config, **kwargs)
        self.dependent_factors = None
        self.base_date = None

        # Cuda related variables to store the state of the device between calculations
        self.shared_memClass = namedtuple('shared_mem',
                                          't_Buffer t_Static_Buffer t_Feed_dict t_Cashflows calc_greeks \
                                          gpus riskneutral precision simulation_batch Report_Currency')

        # prepare the risk factor output matrix . .
        self.static_var = []
        self.static_values = []
        self.static_ofs = OrderedDict()

    def update_factors(self, params, base_date):
        dependent_factors, stochastic_factors, implied_factors, reset_dates, settlement_currencies = \
            self.config.calculate_dependencies(params, base_date, '0d', False)

        # update the time grid
        self.update_time_grid(base_date)

        self.static_factors = OrderedDict()
        for price_factor in dependent_factors:
            try:
                self.static_factors.setdefault(price_factor,
                                               construct_factor(
                                                   price_factor,
                                                   self.config.params['Price Factors']))
            except KeyError as e:
                logging.warning('Price Factor {0} missing in market data file - skipping'.format(e.args))

        self.all_factors = self.static_factors

        # work out distinct tenor sets
        # distinct_tenors = OrderedDict()
        with tf.name_scope("Static_Input"):
            for key, value in self.static_factors.items():
                if key.type not in utils.DimensionLessFactors:
                    # record the offset of this risk factor
                    self.static_ofs.setdefault(key, len(self.static_var))
                    # utils.distinct_tenors(distinct_tenors, key, value)
                    current_val = value.current_value()
                    self.static_values.append(current_val)
                    self.static_var.append(
                        tf.placeholder(self.prec,
                                       shape=len(current_val),
                                       name=utils.check_scope_name(key)))

        # setup the device and allocate memory
        self.__refresh_shared_mem(params['Currency'], params.get('Greeks', 'No') == 'Yes')

        # calculate a reverse lookup for the tenors and store the daycount code
        self.all_tenors = utils.update_tenors(self.base_date, self.all_factors)

        # store the state info
        self.dependent_factors = dependent_factors

    def update_time_grid(self, base_date):
        # setup the scenario and base time grids
        self.time_grid = TimeGrid({base_date}, {base_date}, {base_date})
        self.base_date = base_date
        self.time_grid.set_base_date(base_date)

    def __refresh_shared_mem(self, reporting_currency, calc_greeks):

        # setup the Feed dictionary - the inputs needed to evaluate the graph
        feed = dict(zip(self.static_var, self.static_values))
        static_buffer = self.static_var
        # name of the base currency
        base_currency = 'Static_Input/FxRate.{}'.format(
            self.config.params['System Parameters']['Base_Currency']),

        # now decide what we want to calculate greeks with respect to
        all_vars_concat = None
        if calc_greeks:
            all_vars_concat = [x for x in self.static_var if not x.name.startswith(
                base_currency)]
            self.make_factor_index(self.static_var)

        # allocate memory on the device
        with tf.name_scope('Setup'):
            self.shared_mem = self.shared_memClass(
                t_Buffer={},
                t_Static_Buffer=static_buffer,
                t_Feed_dict=feed,
                t_Cashflows=None,
                calc_greeks=all_vars_concat,
                gpus=[x.name for x in device_lib.list_local_devices() if x.device_type == 'GPU'],
                riskneutral=False,
                precision=self.prec,
                simulation_batch=1,
                Report_Currency=get_fxrate_factor(utils.check_rate_name(reporting_currency),
                                                  self.static_ofs, {})
            )

    def feed(self, static_vars={}):

        def check_prices(sess, feed, n, parent=[]):

            def format_row(deal, data, val, greeks):
                data['Deal Currency'] = deal.Factor_dep.get(
                    'Local_Currency', deal.Instrument.field.get('Currency'))
                for k, v in val.items():
                    if k.startswith('Greeks'):
                        greeks.setdefault(k, []).append(
                            self.gradients_as_df(v, header=deal.Instrument.field.get('Reference')))
                    else:
                        data[k] = float(v)
                # update any tags
                if deal.Instrument.field.get('Tags') is not None:
                    data.update(dict(zip(tag_titles, deal.Instrument.field['Tags'][0].split(','))))

            block = []
            greeks = {}
            for sub_struct in n.sub_structures:
                data = OrderedDict(parent + [(field, sub_struct.obj.Instrument.field.get(field, 'Root'))
                                             for field in ['Reference', 'Object']])
                format_row(sub_struct.obj, data, sess.run(sub_struct.obj.Calc_res, feed), greeks)
                block.append(data)
                sub_block, sub_greeks = check_prices(
                    sess, feed, sub_struct, parent + [('Parent' + str(len(parent)), data['Reference'])])
                block.extend(sub_block)
                # aggregate the sub structure greeks
                for k, v in sub_greeks.items():
                    greeks.setdefault(k, []).extend(v)

            valuations = sess.run([deal.Calc_res for deal in n.dependencies], feed)
            for deal, val in zip(n.dependencies, valuations):
                data = OrderedDict(parent + [(field, deal.Instrument.field.get(field, '?'))
                                             for field in ['Reference', 'Object']])
                format_row(deal, data, val, greeks)
                block.append(data)

            return block, greeks

        # update the feed
        feed = self.shared_mem.t_Feed_dict.copy()
        # copy the new params
        feed.update({self.graph.get_tensor_by_name('Static_Input/{0}:0'.format(k)): v for
                     k, v in static_vars.items()})

        # record the cuda execution stats
        self.calc_stats['Tensor_Execution_Time'] = time.clock()

        # clear the output
        self.output = {}
        # load any tag titles
        tag_titles = self.config.deals['Attributes'].get('Tag_Titles', '').split(',')
        # setup the config
        config = tf.ConfigProto(allow_soft_placement=True)

        with tf.Session(graph=self.graph, config=config) as sess:
            mtm, greeks = check_prices(sess, feed, self.netting_sets)

        self.output['mtm'] = pd.DataFrame(mtm)
        for greek_name, greek_val in greeks.items():
            # this guarantees that the multiindex is uniquely defined when we write it out
            if greek_name == 'Greeks_Second':
                summary = pd.concat(greek_val, axis=1).groupby(level=[0, 1, 2, 3, 4], axis=1).sum()
            elif greek_name == 'Greeks_First':
                summary = pd.concat(greek_val, axis=1).groupby(level=0, axis=1).sum()
            else:
                raise Exception('Unknown Greek requested', greek_name)
            self.output.setdefault(greek_name, summary)

        self.calc_stats['Tensor_Execution_Time'] = time.clock() - self.calc_stats['Tensor_Execution_Time']

        return self.output

    def execute(self, params, feedgraph=True):
        # get the rundate
        base_date = pd.Timestamp(params['Run_Date'])
        # reset the default graph
        tf.reset_default_graph()
        # create a TF graph to represent this calculation
        self.graph = tf.Graph()
        # store the params
        self.params = params
        # set the precision in the utils module - Nasty hack - Need to fix this - TODO
        utils.Default_Precision = self.prec
        # build the TF graph
        with self.graph.as_default() as g:
            self.update_factors(params, base_date)

            self.calc_stats['Deal_Setup_Time'] = time.clock()
            self.netting_sets = DealStructure(Aggregation('root'), deal_level_mtm=True)
            self.set_deal_structures(self.config.deals['Deals']['Children'], self.netting_sets, 1, 1,
                                     deal_level_mtm=True)

            # record the (pure python) dependency setup time
            self.calc_stats['Deal_Setup_Time'] = time.clock() - self.calc_stats['Deal_Setup_Time']

            self.calc_stats['Graph_Setup_Time'] = time.clock()

            # now ask the netting set to construct each deal - no looping required (just 1 timepoint)
            mtm = self.netting_sets.resolve_structure(self.shared_mem, self.time_grid)

            # record the graph loading time
            self.calc_stats['Graph_Setup_Time'] = time.clock() - self.calc_stats['Graph_Setup_Time']

        # return a dictionary of output
        return {'Netting': self.netting_sets, 'Stats': self.calc_stats,
                'Results': self.feed() if feedgraph else None}


def construct_calculation(calc_type, config, **kwargs):
    return globals().get(calc_type)(config, **kwargs)


if __name__ == '__main__':
    pass



import os
import sys
import json
import glob
import traceback
import pandas as pd
import numpy as np

from collections import defaultdict
from multiprocessing import Process, Queue, Lock


class CustomJsonEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return_value = {'.array': obj.tolist()}
        return return_value


def CustomJsonDecoder(dct):
    if '.array' in dct:
        return np.array(dct['.array'])
    return dct


def merge_profiles(id, merge_data, rundate, outputdir):
    pivots = []

    # remove any existing totals
    for prev_totals in glob.glob(outputdir + os.sep + 'Tagged_{}_Total_{}.csv'.format(rundate, id)):
        os.remove(prev_totals)

    pivot_labels = ['Reference', 'Acquirer', 'Portfolio', 'Instrument', 'SalesTeam']
    if 'Factor' in merge_data.values()[0]:
        pivot_labels.insert(1, merge_data.values()[0]['Factor'])

    for name, data in merge_data.items():
        # details for indexing
        profile = pd.read_csv(
            outputdir + os.sep + 'Tagged_{0}_{1}_{2}_{3}.csv'.format(rundate, data['currency'], data['csa'], name))
        position = pd.pivot_table(profile, values=profile.columns[4 + len(data):].tolist(), index=pivot_labels,
                                  aggfunc=np.sum)
        position['Netting'] = name
        position['CSA'] = data['csa']
        position['Currency'] = data['currency']
        pivots.append(position)

    # write it out
    filename = outputdir + os.sep + 'Tagged_{}_Total_{}.csv'.format(rundate, id)
    pd.concat(pivots, axis=0, join='outer').to_csv(filename)
    return ('Merge', filename)


class JOB(object):
    def __init__(self, cx, rundate, input_path, outputdir, netting_set, stats, log,
                 time_grid='0d 2d 1w(1w) 3m(1m) 2y(3m)'):
        self.cx = cx
        self.rundate = rundate
        self.input_path = input_path
        self.outputdir = outputdir
        self.netting_set = netting_set
        self.stats = stats
        self.logger = log
        # load trade
        self.cx.parse_json(os.path.join(self.input_path, self.rundate, self.netting_set))
        # default cmc parameters - note - fix this to load up all the parameters from cx.deals['Calculation']
        self.params = {'calc_name': ('cmc', 'calc1'),
                       'Time_grid': str(self.cx.deals['Calculation']['Base_Time_Grid']),
                       'Run_Date': self.cx.deals['Calculation']['Base_Date'].strftime('%Y-%m-%d'),
                       'Currency': self.cx.deals['Calculation']['Currency'],
                       'Deflation_Interest_Rate': self.cx.deals['Calculation']['Deflation_Interest_Rate'],
                       'Simulation_Batches': 10,
                       'Batch_Size': 512,
                       'Random_Seed': 5126,
                       'Calc_Scenarios': 'No',
                       'Debug': 'No',
                       'NoModel': 'Constant',
                       'Partition': 'None',
                       'Generate_Slideshow': 'No',
                       'PFE_Recon_File': ''}
        # get the netting set
        self.ns = self.cx.deals['Deals']['Children'][0]['instrument'].field
        # get the agreement currency
        self.agreement_currency = self.ns.get('Agreement_Currency', 'ZAR')
        # get the balance currency
        self.balance_currency = self.ns.get('Balance_Currency', self.agreement_currency)

    def perform_calc(self):
        # should be loaded correctly
        from calculation import construct_calculation
        # create the calculation    
        calc = construct_calculation('Credit_Monte_Carlo', self.cx)
        # call the custom calc
        if self.valid():
            self.run_calc(calc)

    def valid(self):
        return False

    def get_filename(self, csa_stat):
        # filename to write the results to
        return 'Tagged_{0}_{1}_{2}_{3}.csv'.format(self.rundate, self.balance_currency, csa_stat, self.netting_set)

    def run_calc(self, calc):
        pass


class PFE(JOB):
    def __init__(self, cx, rundate, input_path, outputdir, netting_set, stats, log):
        super(PFE, self).__init__(cx, rundate, input_path, outputdir, netting_set, stats, log)
        # set the OIS cashflow flag to speed up prime linked swaps
        self.cx.params['Valuation Configuration']['CFFloatingInterestListDeal']['OIS_Cashflow_Group_Size'] = 1

    def valid(self):
        if not self.cx.deals['Deals']['Children'][0]['Children']:
            return False
        else:
            return True

    def run_calc(self, calc):
        deals = defaultdict(int)
        for deal in self.cx.deals['Deals']['Children'][0]['Children']:
            deals[deal['instrument'].field['Object']] += 1

        instruments_in_file = ', '.join(deals)
        # turn on dynamic scenarios (more accurate)
        self.params['Dynamic_Scenario_Dates'] = 'Yes'
        recon_file = os.path.join(self.input_path, self.rundate, self.netting_set[:-4] + 'csv')

        if os.path.isfile(recon_file):
            # load up the file to compare to - csv
            aa_output = pd.read_csv(recon_file, index_col=0)
            # change the currency to match the output from adaptiv
            self.params['Currency'] = aa_output.index.name
            try:
                # run the PFE
                n = calc.execute(self.params)
                exposure = n['Results']['mtm'].clip(0.0, np.inf).astype(np.float64)
                dates = np.array(sorted(calc.time_grid.mtm_dates))[
                    calc.netting_sets.sub_structures[0].obj.Time_dep.deal_time_grid]
                pfe = pd.DataFrame(index=sorted([x.strftime('%Y-%m-%d') for x in dates]),
                                   data={'My_PFE': np.percentile(exposure, 95, axis=1),
                                         'My_EE': exposure.mean(axis=1)})

                result = pd.concat([aa_output, pfe], axis=1)
                self.stats.setdefault('Stats', {})[self.netting_set] = calc.calc_stats
                result['PFE_ABS_ERROR'] = (result['PFE'] - result['My_PFE'])
                result['PFE_REL_ERROR'] = result['PFE_ABS_ERROR'] / result['PFE']
                if ((np.abs(result['PFE_ABS_ERROR']) > 10000) & (np.abs(result['PFE_REL_ERROR']) > 0.1)).any():
                    self.stats['Stats'][self.netting_set]['PFE_ABS_ERROR'] = np.abs(result['PFE_ABS_ERROR']).max()
                    self.stats['Stats'][self.netting_set]['PFE_REL_ERROR'] = np.abs(result['PFE_REL_ERROR']).max()
                if np.abs(result['PFE_REL_ERROR'][0]) > 0.01:
                    self.stats['Stats'][self.netting_set]['Mismatch on Day 0'] = np.abs(result['PFE_REL_ERROR'][0])
                self.stats['Stats'][self.netting_set].update({'instruments': instruments_in_file})
                result.to_csv(self.outputdir + os.sep + 'RECON_' + self.netting_set[:-4] + 'csv')
            except:
                self.logger(self.netting_set, 'Error: PFE did not complete')
        else:
            self.logger(self.netting_set, 'Error: could not find Adaptiv Output')


class CVA_GRAD(JOB):
    def __init__(self, cx, rundate, input_path, outputdir, netting_set, stats, log):
        super(CVA_GRAD, self).__init__(cx, rundate, input_path, outputdir, netting_set, stats, log)
        # load up a calendar (for theta)
        self.business_day = cx.holidays['Johannesburg']['businessday']
        # set the OIS cashflow flag to speed up prime linked swaps
        self.cx.params['Valuation Configuration']['CFFloatingInterestListDeal']['OIS_Cashflow_Group_Size'] = 1

    def valid(self):
        if not self.cx.deals['Deals']['Children'][0]['Children']:
            return False
        else:
            return True

    def run_calc(self, calc):
        from calculation import construct_calculation

        filename = 'CVA_' + self.params['Run_Date'] + '_' + self.cx.deals['Attributes']['Reference'] + '.csv'

        # load up CVA calc params
        if self.cx.deals['Deals']['Children'][0]['instrument'].field.get('Collateralized') == 'True':
            self.logger(self.netting_set, 'is collateralized')
            # turn on dynamic scenarios (more accurate)
            self.params['Dynamic_Scenario_Dates'] = 'Yes'
            self.params['Simulation_Batches'] = 20
            self.params['Batch_Size'] = 256
        else:
            self.params['Dynamic_Scenario_Dates'] = 'No'
            self.params['Simulation_Batches'] = 10
            self.params['Batch_Size'] = 512
            self.logger(self.netting_set, 'is uncollateralized')

        # get the calculation parameters for CVA
        cva_sect = self.cx.deals['Calculation']['Credit_Valuation_Adjustment']

        # update the params
        self.params['CVA'] = {'Counterparty': cva_sect['Counterparty'],
                              'Deflate_Stochastically': cva_sect['Deflate_Stochastically'],
                              'Stochastic_Hazard': cva_sect['Stochastic_Hazard_Rates']}

        # work out the next business day
        next_day = self.business_day.rollforward(
            pd.Timestamp(self.params['Run_Date']) + pd.offsets.Day(1)).strftime('%Y-%m-%d')
        
        if os.path.isfile( os.path.join(self.outputdir, 'Greeks', filename) ):
            self.logger(self.netting_set, 'Warning: skipping CVA gradient calc as file already exists')
            self.params['CVA']['Gradient'] = 'No'
            out = calc.execute(self.params)
            stats = out['Stats']
            # store the CVA as part of the stats
            out['Stats'].update({'CVA': out['Results']['cva'], 'Currency': self.params['Currency']})
            # now calc theta
            self.params.update({'Run_Date': next_day})
            calc = construct_calculation('Credit_Monte_Carlo', self.cx)
            theta = calc.execute(self.params)
            out['Stats'].update({'CVA_Theta': theta['Results']['cva']})
            # store cva
            self.stats.setdefault('Stats', {})[self.netting_set] = out['Stats']
        else:
            self.params['CVA']['Gradient'] = 'Yes'
            att = 0
            while att <= 3:
                try:
                    out = calc.execute(self.params)
                    att = 10
                except Exception as e:
                    exc_type, exc_value, exc_traceback = sys.exc_info()
                    traceback.print_exception(exc_type, exc_value, exc_traceback,
                                              limit=2, file=sys.stdout)
                    att += 1
                    self.params['Simulation_Batches'] *= 2
                    self.params['Batch_Size'] //= 2
                    self.logger(self.netting_set, 'Exception: ' + str(e.args))
                    self.logger(self.netting_set, 'Halving memory to {0} and Doubling Batches to {1}'.format(
                        self.params['Batch_Size'], self.params['Simulation_Batches']))

                    # create a new calculation
                    calc = construct_calculation('Credit_Monte_Carlo', self.cx)
                else:
                    stats = out['Stats']
                    grad_cva = calc.gradients_as_df(out['Results']['grad_cva']).rename(
                        columns={'Gradient': self.cx.deals['Attributes']['Reference']})
                    # store the CVA as part of the stats
                    out['Stats'].update({'CVA': out['Results']['cva'], 'Currency': self.params['Currency']})
                    # write the grad                    
                    grad_cva.to_csv(os.path.join(self.outputdir, 'Greeks', filename))
                    # now calc theta
                    self.params.update({'Run_Date': next_day, 'Gradient': 'No'})
                    calc = construct_calculation('Credit_Monte_Carlo', self.cx)
                    theta = calc.execute(self.params)
                    out['Stats'].update({'CVA_Theta': theta['Results']['cva']})
                    # store it
                    self.stats.setdefault('Stats', {})[self.netting_set] = out['Stats']


class CVA(JOB):
    def __init__(self, cx, rundate, input_path, outputdir, netting_set, stats, log):
        super(CVA, self).__init__(cx, rundate, input_path, outputdir, netting_set, stats, log)
        # load up a calendar (for theta)
        self.business_day = cx.holidays['Johannesburg']['businessday']
        # set the OIS cashflow flag to speed up prime linked swaps
        self.cx.params['Valuation Configuration']['CFFloatingInterestListDeal']['OIS_Cashflow_Group_Size'] = 1

    def valid(self):
        if not self.cx.deals['Deals']['Children'][0]['Children']:
            return False
        else:
            return True

    def run_calc(self, calc):
        from calculation import construct_calculation

        # load up CVA calc params
        if self.cx.deals['Deals']['Children'][0]['instrument'].field.get('Collateralized') == 'True':
            self.logger(self.netting_set, 'is collateralized')
            # turn on dynamic scenarios (more accurate)
            self.params['Dynamic_Scenario_Dates'] = 'Yes'
            self.params['Simulation_Batches'] = 20
            self.params['Batch_Size'] = 256
        else:
            self.params['Dynamic_Scenario_Dates'] = 'No'
            self.params['Simulation_Batches'] = 10
            self.params['Batch_Size'] = 512
            self.logger(self.netting_set, 'is uncollateralized')

        # get the calculation parameters for CVA
        cva_sect = self.cx.deals['Calculation']['Credit_Valuation_Adjustment']

        # update the params
        self.params['CVA'] = {'Counterparty': cva_sect['Counterparty'],
                              'Deflate_Stochastically': cva_sect['Deflate_Stochastically'],
                              'Stochastic_Hazard': cva_sect['Stochastic_Hazard_Rates'],
                              'Gradient': 'No'}
        att = 0
        while att <= 3:
            try:
                out = calc.execute(self.params)
                att = 10
            except Exception as e:
                exc_type, exc_value, exc_traceback = sys.exc_info()
                traceback.print_exception(exc_type, exc_value, exc_traceback,
                                          limit=2, file=sys.stdout)
                att += 1
                self.params['Simulation_Batches'] *= 2
                self.params['Batch_Size'] //= 2
                self.logger(self.netting_set, 'Exception: ' + str(e.args))
                self.logger(self.netting_set, 'Halving memory to {0} and Doubling Batches to {1}'.format(
                    self.params['Batch_Size'], self.params['Simulation_Batches']))

                # create a new calculation
                calc = construct_calculation('Credit_Monte_Carlo', self.cx)
            else:
                stats = out['Stats']
                # store the CVA as part of the stats
                out['Stats'].update({'CVA': out['Results']['cva'], 'Currency': self.params['Currency']})
                # store it
                self.stats.setdefault('Stats', {})[self.netting_set] = out['Stats']


class COLLVA(JOB):
    def __init__(self, cx, rundate, input_path, outputdir, netting_set, stats, log):
        self.rundate = rundate
        self.input_path = input_path
        self.outputdir = outputdir
        self.netting_set = os.path.splitext(netting_set)[0] + '.json'
        self.stats = stats
        self.logger = log
        self.params = {'calc_name': ('cmc', 'calc1'),
                       'Time_grid': '0d 2d 1w(1w) 1m(1m) 3m(3m) 1y(1y)',
                       'Run_Date': rundate, 'Currency': 'ZAR', 'Random_Seed': 5126,
                       'Calc_Scenarios': 'No', 'Dynamic_Scenario_Dates': 'Yes',
                       'Debug': 'No', 'NoModel': 'Constant', 'Partition': 'None',
                       'Generate_Slideshow': 'No', 'PFE_Recon_File': ''}

        from adaptiv import AdaptivContext
        self.cx = AdaptivContext()
        old_cx = AdaptivContext()
        # load marketdata
        old_cx.parse_json(os.path.join(self.input_path, rundate, 'MarketData.json'))
        # load up the CVA marketdata file
        self.cx.parse_json(os.path.join(self.input_path, rundate, 'MarketDataCVA.json'))
        # update the cva data with the arena data
        self.cx.params['Price Factors'].update(old_cx.params['Price Factors'])
        # load trade
        self.cx.parse_json(os.path.join(self.input_path, self.rundate, self.netting_set))
        # get the netting set
        self.ns = self.cx.deals['Deals']['Children'][0]['instrument'].field
        # get the agreement currency
        self.agreement_currency = self.ns.get('Agreement_Currency', 'ZAR')
        # get the balance currency
        self.balance_currency = self.ns.get('Balance_Currency', self.agreement_currency)
        # set the OIS cashflow flag to speed up prime linked swaps
        self.cx.params['Valuation Configuration']['CFFloatingInterestListDeal']['OIS_Cashflow_Group_Size'] = 1
        from utils import Curve
        # change the currency
        self.params['Currency'] = self.agreement_currency

        def makeflatcurve(curr, bps, tenor=30):
            return {'Currency': curr, 'Curve': Curve([], [[0, bps * 0.01 * 0.01], [tenor, bps * 0.01 * 0.01]]),
                    'Day_Count': 'ACT_365',
                    'Property_Aliases': None, 'Sub_Type': 'None'}

        if self.agreement_currency == 'ZAR':
            self.cx.params['Price Factors']['InterestRate.ZAR-SWAP.OIS'] = makeflatcurve('ZAR', -15)
            self.cx.params['Price Factors']['InterestRate.ZAR-SWAP.FUNDING'] = makeflatcurve('ZAR', 10)
            self.cx.deals['Deals']['Children'][0]['instrument'].field['Collateral_Assets']['Cash_Collateral'][0][
                'Collateral_Rate'] = 'ZAR-SWAP.OIS'
            self.cx.deals['Deals']['Children'][0]['instrument'].field['Collateral_Assets']['Cash_Collateral'][0][
                'Funding_Rate'] = 'ZAR-SWAP.FUNDING'
        else:
            self.cx.params['Price Factors']['InterestRate.USD-LIBOR-3M.FUNDING'] = makeflatcurve('USD', 65)
            self.cx.deals['Deals']['Children'][0]['instrument'].field['Collateral_Assets']['Cash_Collateral'][0][
                'Collateral_Rate'] = 'USD-OIS'
            self.cx.deals['Deals']['Children'][0]['instrument'].field['Collateral_Assets']['Cash_Collateral'][0][
                'Funding_Rate'] = 'USD-LIBOR-3M.FUNDING'

    def valid(self):
        if not self.cx.deals['Deals']['Children'][0]['Children'] or self.cx.deals[
            'Deals']['Children'][0]['instrument'].field.get('Collateralized', 'False') == 'False':
            return False
        else:
            return True

    def run_calc(self, calc):
        filename = 'COLLVA_' + self.params['Run_Date'] + '_' + self.crb_default + '.csv'

        if os.path.isfile(os.path.join(self.outputdir, 'Greeks', filename)):
            self.logger(self.netting_set, 'Warning: skipping COLLVA calc as file already exists')
        else:
            self.params['CollVA'] = {'Gradient': 'Yes'}

            # make sure the margin period of risk is 10 business days (approx 12 calendar days)
            self.cx.deals['Deals']['Children'][0]['instrument'].field['Liquidation_Period'] = 12.0
            self.params['Simulation_Batches'] = 20
            self.params['Batch_Size'] = 256

            try:
                out = calc.execute(self.params)
            except Exception as e:
                exc_type, exc_value, exc_traceback = sys.exc_info()
                traceback.print_exception(exc_type, exc_value, exc_traceback,
                                          limit=2, file=sys.stdout)

                self.logger(self.netting_set, 'Exception: ' + str(e.args))
            else:
                stats = out['Stats']
                grad_collva = calc.gradients_as_df(out['Results']['grad_collva']).rename(
                    columns={'Gradient': self.cx.deals['Attributes']['Reference']})
                # store the CollVA as part of the stats
                out['Stats'].update({'CollVA': out['Results']['collva'], 'Currency': self.params['Currency']})
                grad_collva.to_csv(os.path.join(self.outputdir, 'Greeks', filename))
                self.stats.setdefault('Stats', {})[self.netting_set] = out['Stats']
                # log the netting set
                self.logger(self.netting_set, 'CollVA calc complete')


class CVADEFAULT(JOB):
    def __init__(self, cx, rundate, input_path, outputdir, netting_set, stats, log):
        self.rundate = rundate
        self.input_path = input_path
        self.outputdir = outputdir
        self.netting_set = os.path.splitext(netting_set)[0] + '.json'
        self.stats = stats
        self.logger = log
        self.params = {'calc_name': ('cmc', 'calc1'),
                       'Time_grid': '0d 2d 1w(1w) 1m(1m) 3m(3m) 1y(1y)',
                       'Run_Date': rundate, 'Currency': 'ZAR', 'Random_Seed': 5126,
                       'Calc_Scenarios': 'No', 'Dynamic_Scenario_Dates': 'Yes',
                       'Debug': 'No', 'NoModel': 'Constant', 'Partition': 'None',
                       'Generate_Slideshow': 'No', 'PFE_Recon_File': ''}

        from adaptiv import AdaptivContext
        self.cx = AdaptivContext()
        old_cx = AdaptivContext()
        # load marketdata
        old_cx.parse_json(os.path.join(self.input_path, rundate, 'MarketData.json'))
        # load up the CVA marketdata file
        self.cx.parse_json(os.path.join(self.input_path, rundate, 'MarketDataCVA.json'))
        # update the cva data with the arena data
        self.cx.params['Price Factors'].update(old_cx.params['Price Factors'])
        # load trade
        self.cx.parse_json(os.path.join(self.input_path, self.rundate, self.netting_set))
        # get the netting set
        self.ns = self.cx.deals['Deals']['Children'][0]['instrument'].field
        # get the agreement currency
        self.agreement_currency = self.ns.get('Agreement_Currency', 'ZAR')
        # get the balance currency
        self.balance_currency = self.ns.get('Balance_Currency', self.agreement_currency)
        # set the OIS cashflow flag to speed up prime linked swaps
        self.cx.params['Valuation Configuration']['CFFloatingInterestListDeal']['OIS_Cashflow_Group_Size'] = 1
        from utils import Curve
        # set the survival curve to a default value
        self.crb_default = self.cx.deals['Attributes']['Reference']
        self.cx.params['Price Factors'].setdefault('SurvivalProb.' + self.crb_default,
                                                   {'Recovery_Rate': 0.5, 'Curve': Curve(
                                                       [],
                                                       [[0.0, 0.0], [.5, .01], [1, .02], [3, .07], [5, .15], [10, .35],
                                                        [20, .71]]), 'Property_Aliases': None})

    def valid(self):
        if not self.cx.deals['Deals']['Children'][0]['Children']:
            return False
        else:
            return True

    def run_calc(self, calc):
        filename = 'CVA_' + self.params['Run_Date'] + '_' + self.crb_default + '.csv'

        if os.path.isfile(os.path.join(self.outputdir, 'Greeks', filename)):
            self.logger(self.netting_set, 'Warning: skipping CVA calc as file already exists')
        else:
            self.params['CVA'] = {'Deflation': self.cx.deals['Calculation'].get('Deflation_Interest_Rate', 'ZAR-SWAP'),
                                  'Gradient': 'Yes'}

            if self.cx.deals['Deals']['Children'][0]['instrument'].field.get('Collateralized') == 'True':
                self.logger(self.netting_set, 'is collateralized')
                # make sure the margin period of risk is 10 business days (approx 12 calendar days)
                self.cx.deals['Deals']['Children'][0]['instrument'].field['Liquidation_Period'] = 12.0
                self.params['Simulation_Batches'] = 20
                self.params['Batch_Size'] = 256
            else:
                self.params['Simulation_Batches'] = 10
                self.params['Batch_Size'] = 512
                self.logger(self.netting_set, 'is uncollateralized')

            # get the calculation parameters for CVA
            default_cva = {'Deflate_Stochastically': 'Yes', 'Stochastic_Hazard_Rates': 'No',
                           'Counterparty': self.crb_default}
            cva_sect = self.cx.deals.get('Calculation', {'Credit_Valuation_Adjustment': default_cva}).get(
                'Credit_Valuation_Adjustment', default_cva)

            # update the params
            self.params['CVA']['Counterparty'] = cva_sect['Counterparty']
            self.params['CVA']['Deflate_Stochastically'] = cva_sect['Deflate_Stochastically']
            self.params['CVA']['Stochastic_Hazard'] = cva_sect['Stochastic_Hazard_Rates']
            # now adjust the survival curve - need intervals of .25 years
            sc = self.cx.params['Price Factors']['SurvivalProb.' + cva_sect['Counterparty']]['Curve'].array.copy()
            tenors = np.arange(0, sc[-1][0], .25)
            self.cx.params['Price Factors']['SurvivalProb.' + cva_sect['Counterparty']]['Curve'].array = np.array(
                list(zip(tenors, np.interp(tenors, sc[:, 0], sc[:, 1]))))

            try:
                out = calc.execute(self.params)
            except Exception as e:
                exc_type, exc_value, exc_traceback = sys.exc_info()
                traceback.print_exception(exc_type, exc_value, exc_traceback,
                                          limit=2, file=sys.stdout)

                self.logger(self.netting_set, 'Exception: ' + str(e.args))
            else:
                stats = out['Stats']
                grad_cva = calc.gradients_as_df(out['Results']['grad_cva']).rename(
                    columns={'Gradient': self.cx.deals['Attributes']['Reference']})
                # store the CVA as part of the stats
                out['Stats'].update({'CVA': out['Results']['cva'], 'Currency': self.params['Currency']})
                grad_cva.to_csv(os.path.join(self.outputdir, 'Greeks', filename))
                self.stats.setdefault('Stats', {})[self.netting_set] = out['Stats']
                # log the netting set
                self.logger(self.netting_set, 'CVA calc complete')


class FVADEFAULT(JOB):
    def __init__(self, cx, rundate, input_path, outputdir, netting_set, stats, log):
        self.rundate = rundate
        self.input_path = input_path
        self.outputdir = outputdir
        self.netting_set = os.path.splitext(netting_set)[0] + '.json'
        self.stats = stats
        self.logger = log
        self.params = {'calc_name': ('cmc', 'calc1'),
                       'Time_grid': '0d 2d 1w(1w) 1m(1m) 3m(3m) 1y(1y)',
                       'Run_Date': rundate, 'Currency': 'USD', 'Random_Seed': 5126,
                       'Calc_Scenarios': 'No', 'Dynamic_Scenario_Dates': 'Yes',
                       'Debug': 'No', 'NoModel': 'Constant', 'Partition': 'None',
                       'Generate_Slideshow': 'No', 'PFE_Recon_File': ''}

        from adaptiv import AdaptivContext
        self.cx = AdaptivContext()
        old_cx = AdaptivContext()
        # load marketdata
        old_cx.parse_json(os.path.join(self.input_path, rundate, 'MarketData.json'))
        # load up the CVA marketdata file
        self.cx.parse_json(os.path.join(self.input_path, rundate, 'MarketDataCVA.json'))
        # update the cva data with the arena data
        self.cx.params['Price Factors'].update(old_cx.params['Price Factors'])
        # load trade
        self.cx.parse_json(os.path.join(self.input_path, self.rundate, self.netting_set))
        # get the netting set
        self.ns = self.cx.deals['Deals']['Children'][0]['instrument'].field
        # get the agreement currency
        self.agreement_currency = self.ns.get('Agreement_Currency', 'USD')
        # get the balance currency
        self.balance_currency = self.ns.get('Balance_Currency', self.agreement_currency)
        # set the OIS cashflow flag to speed up prime linked swaps
        self.cx.params['Valuation Configuration']['CFFloatingInterestListDeal']['OIS_Cashflow_Group_Size'] = 1
        from utils import Curve

        def makeflatcurve(curr, bps, tenor=30):
            return {'Currency': curr, 'Curve': Curve([], [[0, bps * 0.01 * 0.01], [tenor, bps * 0.01 * 0.01]]),
                    'Day_Count': 'ACT_365',
                    'Property_Aliases': None, 'Sub_Type': 'None'}

            # set the survival curve to a default value

        self.crb_default = self.cx.deals['Attributes']['Reference']
        # set up funding curves - ??? 
        # self.cx.params['Price Factors']['InterestRate.ZAR-SWAP.OIS'] = makeflatcurve('ZAR',-15)
        # self.cx.params['Price Factors']['InterestRate.ZAR-SWAP.FUNDING'] = makeflatcurve('ZAR', 10)

        self.cx.params['Price Factors']['InterestRate.USD-LIBOR-3M.FUNDING'] = makeflatcurve('USD', 65)

        self.cx.params['Price Factors'].setdefault('SurvivalProb.' + self.crb_default,
                                                   {'Recovery_Rate': 0.5, 'Curve': Curve(
                                                       [],
                                                       [[0.0, 0.0], [.5, .01], [1, .02], [3, .07], [5, .15], [10, .35],
                                                        [20, .71], [30, 1.0]]), 'Property_Aliases': None})

    def valid(self):
        if not self.cx.deals['Deals']['Children'][0]['Children']:
            return False
        else:
            return True

    def run_calc(self, calc):
        filename = 'FVA_' + self.params['Run_Date'] + '_' + self.crb_default + '.csv'

        if os.path.isfile(os.path.join(self.outputdir, 'Greeks', filename)):
            self.logger(self.netting_set, 'Warning: skipping FVA calc as file already exists')
        else:
            self.params['FVA'] = {'Funding_Interest_Curve': 'USD-LIBOR-3M.FUNDING',
                                  'Risk_Free_Curve': 'USD-OIS',
                                  'Stochastic_Funding': 'Yes', 'Counterparty': self.crb_default, 'Gradient': 'Yes'}

            if self.cx.deals['Deals']['Children'][0]['instrument'].field.get('Collateralized', 'False') != 'True':
                # only calc FVA for uncollateralized counterparties
                self.params['Simulation_Batches'] = 10
                self.params['Batch_Size'] = 512
                self.logger(self.netting_set, 'is uncollateralized')

                try:
                    out = calc.execute(self.params)
                except Exception as e:
                    exc_type, exc_value, exc_traceback = sys.exc_info()
                    traceback.print_exception(exc_type, exc_value, exc_traceback,
                                              limit=2, file=sys.stdout)
                    self.logger(self.netting_set, 'Exception: ' + str(e.args))
                else:
                    stats = out['Stats']
                    if 'grad_fva' in out['Results']:
                        grad_fva = calc.gradients_as_df(out['Results']['grad_fva']).rename(
                            columns={'Gradient': self.cx.deals['Attributes']['Reference']})
                        grad_fva.to_csv(os.path.join(self.outputdir, 'Greeks', filename))
                    # store the FVA as part of the stats
                    out['Stats'].update({'FVA': out['Results']['fva'], 'Currency': self.params['Currency']})
                    self.stats.setdefault('Stats', {})[self.netting_set] = out['Stats']
                    # log the netting set
                    self.logger(self.netting_set, 'FVA calc complete')


class CVAVega(JOB):
    def __init__(self, cx, rundate, input_path, outputdir, netting_set, stats, log):
        self.rundate = rundate
        self.input_path = input_path
        self.outputdir = outputdir
        self.netting_set = os.path.splitext(netting_set)[0] + '.json'
        self.stats = stats
        self.logger = log
        self.params = {'calc_name': ('cmc', 'calc1'),
                       'Time_grid': '0d 2d 1w(1w) 3m(1m) 2y(3m)', 'Run_Date': rundate,
                       'Currency': 'ZAR', 'Random_Seed': 5126,
                       'Calc_Scenarios': 'No',
                       'Dynamic_Scenario_Dates': 'Yes',
                       'Debug': 'No', 'NoModel': 'Constant',
                       'Partition': 'None', 'Generate_Slideshow': 'No',
                       'PFE_Recon_File': ''}

        from adaptiv import AdaptivContext
        self.cx = AdaptivContext()
        self.cx.parse_json(os.path.join(self.input_path, rundate, 'MarketDataCVA.json'))
        self.noshift = AdaptivContext()
        self.noshift.parse_json(os.path.join(self.input_path, rundate, 'MarketDataNoShift.json'))
        self.shift = AdaptivContext()
        self.shift.parse_json(os.path.join(self.input_path, rundate, 'MarketDataShift.json'))

        # update the cva data with the arena data
        self.cx.params['Price Factors'].update(self.noshift.params['Price Factors'])

        self.cx.parse_json(os.path.join(self.input_path, rundate, self.netting_set))
        # get the netting set
        self.ns = self.cx.deals['Deals']['Children'][0]['instrument'].field
        # get the agreement currency
        self.agreement_currency = self.ns.get('Agreement_Currency', 'ZAR')
        # get the balance currency
        self.balance_currency = self.ns.get('Balance_Currency', self.agreement_currency)
        # set the OIS cashflow flag to speed up prime linked swaps
        self.cx.params['Valuation Configuration']['CFFloatingInterestListDeal']['OIS_Cashflow_Group_Size'] = 1
        from utils import Curve
        # set the survival curve to a default value
        self.crb_default = self.cx.deals['Attributes']['Reference']
        self.cx.params['Price Factors'].setdefault('SurvivalProb.' + self.crb_default,
                                                   {'Recovery_Rate': 0.5, 'Curve': Curve(
                                                       [],
                                                       [[0.0, 0.0], [.5, .01], [1, .02], [3, .07], [5, .15], [10, .35],
                                                        [20, .71]]), 'Property_Aliases': None})

    def valid(self):
        if not self.cx.deals['Deals']['Children'][0]['Children']:
            return False
        else:
            return True

    def run_calc(self, calc):
        from calculation import construct_calculation

        self.params['CVA'] = {'Deflation': self.cx.deals['Calculation'].get('Deflation_Interest_Rate', 'ZAR-SWAP'),
                              'Gradient': 'No'}

        if self.cx.deals['Deals']['Children'][0]['instrument'].field.get('Collateralized') == 'True':
            self.logger(self.netting_set, 'is collateralized')
            self.params['Simulation_Batches'] = 20
            self.params['Batch_Size'] = 256
        else:
            self.params['Simulation_Batches'] = 10
            self.params['Batch_Size'] = 512
            self.logger(self.netting_set, 'is uncollateralized')

        # get the calculation parameters for CVA
        default_cva = {'Deflate_Stochastically': 'Yes', 'Stochastic_Hazard_Rates': 'No',
                       'Counterparty': self.crb_default}
        cva_sect = self.cx.deals.get('Calculation', {'Credit_Valuation_Adjustment': default_cva}).get(
            'Credit_Valuation_Adjustment', default_cva)

        # update the params
        self.params['CVA']['Counterparty'] = cva_sect['Counterparty']
        self.params['CVA']['Deflate_Stochastically'] = cva_sect['Deflate_Stochastically']
        self.params['CVA']['Stochastic_Hazard'] = cva_sect['Stochastic_Hazard_Rates']

        try:
            out = calc.execute(self.params)
            self.cx.params['Price Factors'].update(self.shift.params['Price Factors'])
            calcshift = construct_calculation('Credit_Monte_Carlo', self.cx)
            outshift = calcshift.execute(self.params)
        except Exception as e:
            exc_type, exc_value, exc_traceback = sys.exc_info()
            traceback.print_exception(exc_type, exc_value, exc_traceback,
                                      limit=2, file=sys.stdout)

            self.logger(self.netting_set, 'Exception: ' + str(e.args))
        else:
            stats = out['Stats']
            stats.update({'CVA': out['Results']['cva'], 'CVA_Shift': outshift['Results']['cva'],
                          'Currency': self.params['Currency']})
            self.stats.setdefault('Stats', {})[self.netting_set] = stats
            # log the netting set
            self.logger(self.netting_set, 'CVA calc complete NoShift:{}, Shift:{}'.format(
                out['Results']['cva'], outshift['Results']['cva']))


class BaseVal(JOB):
    def __init__(self, cx, rundate, input_path, outputdir, netting_set, stats, log):
        self.cx = cx
        self.rundate = rundate
        self.input_path = input_path
        self.outputdir = outputdir
        self.stats = stats
        self.logger = log
        self.business_day = cx.holidays['Johannesburg']['businessday']
        self.netting_set = os.path.splitext(netting_set)[0] + '.json'
        cx.parse_json(os.path.join(self.input_path, rundate, self.netting_set))
        self.params = {'calc_name': ('cmc', 'calc1'), 'Run_Date': rundate}

    def valid(self):
        if not self.cx.deals['Deals']['Children'][0]['Children']:
            return False
        else:
            return True

    def run_calc(self, calc):
        # needed for theta
        next_day = self.business_day.rollforward(
            pd.Timestamp(self.params['Run_Date']) + pd.offsets.Day(1)).strftime('%Y-%m-%d')
        ending = self.params['Run_Date'] + '_' + self.cx.deals['Attributes']['Reference'] + '.csv'
        filename = 'BaseVal_' + ending
        filename_theta = 'BaseVal_Theta_' + ending
        filename_greeks = 'BaseVal_Delta_' + ending
        filename_greeks_second = 'BaseVal_Gamma_' + ending

        if os.path.isfile(os.path.join(self.outputdir, 'Greeks', filename)):
            self.logger(self.netting_set, 'Warning: skipping BaseVal calc as file already exists')
        else:
            # should be loaded correctly
            from calculation import construct_calculation
            calc = construct_calculation('Base_Revaluation', self.cx, prec=np.float64)
            self.params.update({'Currency': 'ZAR', 'Greeks': 'No'})
            base_val = calc.execute(self.params)
            base_val['Results']['mtm'].to_csv(os.path.join(self.outputdir, 'Greeks', filename))
            try:
                self.params.update({'Run_Date': next_day, 'Greeks': 'Yes'})
                calc = construct_calculation('Base_Revaluation', self.cx, prec=np.float64)
                out = calc.execute(self.params)
            except Exception as e:
                exc_type, exc_value, exc_traceback = sys.exc_info()
                traceback.print_exception(exc_type, exc_value, exc_traceback, limit=2, file=sys.stdout)
                self.logger(self.netting_set, 'Failed to baseval')
            else:
                self.stats.setdefault('Stats', {})[self.netting_set] = out['Stats']
                out['Results']['mtm'].to_csv(os.path.join(self.outputdir, 'Greeks', filename_theta))
                out['Results']['Greeks_First'].to_csv(os.path.join(self.outputdir, 'Greeks', filename_greeks))
                out['Results']['Greeks_Second'].to_csv(os.path.join(self.outputdir, 'Greeks', filename_greeks_second))


def work(id, lock, queue, results, job, rundate, input_path, calendar, outputdir):
    def log(netting_set, msg):
        lock.acquire()
        print('JOB %s: ' % id)
        print('Netting set {0}: {1}'.format(netting_set, msg))
        lock.release()

    # set the visible GPU
    os.environ['CUDA_VISIBLE_DEVICES'] = str(id)
    # set the log level
    os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

    # now load the cuda context
    from adaptiv import AdaptivContext

    # create the crstal context
    cx = AdaptivContext()
    # load calendars
    cx.parse_calendar_file(calendar)
    # load marketdata
    cx.parse_json(os.path.join(input_path, rundate, 'MarketData.json'))

    if os.path.isfile(os.path.join(input_path, rundate, 'CVAMarketData_Calibrated_New.json')):
        cx_new = AdaptivContext()
        cx_new.parse_json(os.path.join(input_path, rundate, 'CVAMarketData_Calibrated_New.json'))
        log("Parent", "Overriding Calibration")
        for factor in [x for x in cx_new.params['Price Factors'].keys()
                       if x.startswith('HullWhite2FactorModelParameters')]:
            # override it
            cx.params['Price Factors'][factor] = cx_new.params['Price Factors'][factor]

    # log results
    logs = {}

    while True:
        # get the task
        task = queue.get()
        if task is None:
            break

        obj = globals().get(job)(cx, rundate, input_path, outputdir, task, logs, log)
        obj.perform_calc()

    # empty the queue
    queue.put(None)
    # get ready to send the results
    result = []

    # write out the logs
    if 'Stats' in logs:
        stats_file = os.path.join(outputdir, 'Stats', '{0}_Stats_{1}_JOB_{2}.csv'.format(job, rundate, id))
        pd.DataFrame(data=logs['Stats']).T.to_csv(stats_file)
        result.append(('Stats', stats_file))
    if 'Merge' in logs:
        result.append(merge_profiles(id, logs['Merge'], rundate, outputdir))
    if 'CSA' in logs:
        csa_file = os.path.join(outputdir, 'CSA', '{0}_{1}_CSA_JOB_{2}.csv'.format(job, rundate, id))
        pd.DataFrame(logs['CSA']).T.to_csv(csa_file)
        result.append(('CSA', csa_file))
    results.put(result)


class Parent:
    def __init__(self, num_jobs):
        self.queue = Queue()
        self.results = Queue()
        self.lock = Lock()
        self.NUMBER_OF_PROCESSES = num_jobs

    def start(self, job, rundate, input_path, calendar, outputdir, wildcard='CrB*.json'):
        print("starting {0} workers in {1}".format(self.NUMBER_OF_PROCESSES, input_path))
        self.workers = [Process(target=work, args=(
            i, self.lock, self.queue, self.results, job, rundate, input_path, calendar, outputdir))
                        for i in range(self.NUMBER_OF_PROCESSES)]

        # start all children
        for w in self.workers:
            w.start()

        crbs = map(lambda x: os.path.split(x)[-1],
                   sorted(glob.glob(os.path.join(input_path, rundate, wildcard)), key=os.path.getsize)[::-1])

        # load the crbs on the queue
        for netting_set in crbs:
            self.queue.put(netting_set)

        self.queue.put(None)

        # now collate the results
        post_processing = []
        for i in range(self.NUMBER_OF_PROCESSES):
            post_processing.append(self.results.get())

        # close the queue
        self.queue.close()
        self.results.close()

        # join to this proces
        for i in range(self.NUMBER_OF_PROCESSES):
            self.workers[i].join()

        post_results = {'Stats': [], 'CSA': [], 'Merge': []}
        for output in post_processing:
            data = dict(output)
            for k, v in data.items():
                post_results[k].append(pd.read_csv(v, index_col=0))

        # write out the combined data
        for k, v in post_results.items():
            if v:
                out_path = os.path.join(outputdir, k, '{0}_{1}_{2}_Total.csv'.format(job, k, rundate))
                pd.concat(v).to_csv(out_path)


if __name__ == '__main__':
    import argparse

    jobs = [cls.__name__ for cls in globals().values() if
            isinstance(cls, type) and hasattr(cls, 'valid') and cls.__name__ != 'JOB']
    # setup the arguments
    parser = argparse.ArgumentParser(description='Run a riskflow batch on a directory of .json netting sets.')
    parser.add_argument('num_jobs', type=int, help='the number of gpu\'s to use (if available) else cpu\'s')
    parser.add_argument('job', type=str, help='the job name', choices=jobs)
    parser.add_argument('rundate', type=str, help='batch rundate')
    parser.add_argument('input_path', type=str,
                        help='directory containing the input files (note that the rundate is assumed to be a directory within)')
    parser.add_argument('calendar_file', type=str, help='calendar file to use')
    parser.add_argument('output_path', type=str, help='output directory')
    parser.add_argument('filename', type=str, help='filename(s) in input_path to run - wildcards allowed')

    # get the arguments
    args = parser.parse_args()

    # print args.rundate, args.input_path, args.calendar_file, args.output_path
    Parent(args.num_jobs).start(args.job, args.rundate, args.input_path, args.calendar_file, args.output_path,
                                args.filename)



#########################################################################
# Copyright (C) 2016-2017  Shuaib Osman (sosman@investec.co.za)
# This file is part of RiskFlow.
#
# RiskFlow is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# RiskFlow is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RiskFlow.  If not, see <http://www.gnu.org/licenses/>.
#########################################################################

# import standard libraries
import json
import calendar
import operator
import itertools
import logging

# import parsing libraries
from typing import Dict, Any, Union

from pyparsing import Literal, Word, nums, OneOrMore, delimitedList, oneOf, Optional, Group
from xml.etree.ElementTree import ElementTree, Element
from functools import reduce

# useful types
from collections import OrderedDict

# dependent libs
import pandas as pd
import numpy as np

# import crstal libraries
import utils
from instruments import construct_instrument, Deal
from bootstrappers import construct_bootstrapper
from stochasticprocess import construct_calibration_config, construct_process

# define datetime routines
Timestamp = pd.Timestamp
DateOffset = pd.DateOffset


def get_grid_grammar():
    """
    Contains the grammar definition rules for parsing the grid data
    """

    def push_int(strg, loc, toks):
        return int(toks[0])

    def push_single_period(strg, loc, toks):
        return Context.offset_lookup[toks[1]], toks[0]

    def push_period(strg, loc, toks):
        ofs = dict(toks.asList())
        return DateOffset(**ofs)

    def push_date_grid(strg, loc, toks):
        return toks[0][0] if len(toks) == 1 else utils.Offsets(toks.asList())

    lpar = Literal("(").suppress()
    rpar = Literal(")").suppress()
    decimal = Literal(".")

    integer = (Word("+-" + nums, nums) + ~decimal).setName('int').setParseAction(push_int)
    single_period = (integer + oneOf(['D', 'M', 'Y', 'W'], caseless=True)).setName('single_period').setParseAction(
        push_single_period)
    period = OneOrMore(single_period).setName('period').setParseAction(push_period)
    grid = delimitedList(Group(period + Optional(lpar + period + rpar)),
                         delim=' ').leaveWhitespace().setParseAction(push_date_grid)

    return grid


class ModelParams(object):
    def __init__(self, state=None):
        # valid risk factor subtypes (only Interest rates at the moment)
        self.valid_subtype = {'BasisSpread': 'BasisSpread'}
        # these models need these additional price factors
        self.implied_models = {'GBMAssetPriceTSModelImplied': 'GBMTSImpliedParameters',
                               'HullWhite2FactorImpliedInterestRateModel': 'HullWhite2FactorModelParameters'}

        self.modeldefaults = OrderedDict()
        self.modelfilters = OrderedDict()
        if state:
            defaults, filters = state
            for factor, model in defaults.items():
                self.append(factor, (), model)
            for factor, mappings in filters.items():
                for condition, model in mappings:
                    self.append(factor, tuple(condition), model)

    def append(self, price_factor, price_filter, stoch_proc):
        if price_filter == ():
            self.modeldefaults.setdefault(price_factor, stoch_proc)
        else:
            self.modelfilters.setdefault(price_factor, []).append((price_filter, stoch_proc))

    def write_to_file(self, file_handle):
        # write out the defaults first, then any filters that apply
        filters_written = set()
        for factor, model_name in self.modeldefaults.items():
            rules = self.modelfilters.get(factor)
            if rules:
                for (attrib, value), model in rules:
                    file_handle.write('{0}={1} where {2} = "{3}"\n'.format(factor, model, attrib, value))
                filters_written.add(factor)
            file_handle.write('{0}={1}\n'.format(factor, model_name))

        # make sure we write all the filters out
        for factor, rules in self.modelfilters.items():
            if factor not in filters_written:
                for (attrib, value), model in rules:
                    file_handle.write('{0}={1} where {2} = "{3}"\n'.format(factor, model, attrib, value))

    def additional_factors(self, model, factor):
        add_factor = self.implied_models.get(model)
        return utils.Factor(add_factor, factor.name) if add_factor else None

    def search(self, factor, actual_factor):
        """
        :param factor: Riskfactor of type utils.Factor
        :param actual_factor: corresponding dictionary of parameters for the factor loaded from a marketdata context
        :return: the model associated with the factor (taking any overrides into account)
        """
        price_factor_type = factor.type + self.valid_subtype.get(actual_factor.get('Sub_Type'), '')
        # look for a filter rule
        rule = self.modelfilters.get(price_factor_type)
        if rule:
            factor_attribs = dict(actual_factor, **{'ID': '.'.join(factor.name)})
            for (attrib, value), model in rule:
                if factor_attribs.get(attrib.strip()) == value.strip():
                    return model
        return self.modeldefaults.get(price_factor_type)


class CustomJsonEncoder(json.JSONEncoder):
    def default(self, obj):
        return_value = {'.Unknown': str(type(obj))}
        if isinstance(obj, utils.Curve):
            return_value = {'.Curve': {'meta': obj.meta, 'data': obj.array.tolist()}}
        elif isinstance(obj, Deal):
            return_value = {'.Deal': obj.field}
        elif isinstance(obj, ModelParams):
            return_value = {'.ModelParams': {'modeldefaults': obj.modeldefaults, 'modelfilters': obj.modelfilters}}
        elif isinstance(obj, utils.Descriptor):
            return_value = {'.Descriptor': obj.data}
        elif isinstance(obj, utils.Percent):
            return_value = {'.Percent': 100.0 * obj.amount}
        elif isinstance(obj, utils.Basis):
            return_value = {'.Basis': 10000.0 * obj.amount}
        elif isinstance(obj, utils.Offsets):
            return_value = {'.Offsets': obj.data}
        elif isinstance(obj, utils.CreditSupportList):
            return_value = {'.CreditSupportList': [[rating, value] for rating, value in obj.data.items()]}
        elif isinstance(obj, utils.DateEqualList):
            return_value = {
                '.DateEqualList': [[date.strftime('%Y-%m-%d')] + list(value) for date, value in obj.data.items()]}
        elif isinstance(obj, utils.DateList):
            return_value = {'.DateList': [[date.strftime('%Y-%m-%d'), value] for date, value in obj.data.items()]}
        elif isinstance(obj, DateOffset):
            return_value = {'.DateOffset': obj.kwds}
        elif isinstance(obj, Timestamp):
            return_value = {'.Timestamp': obj.strftime("%Y-%m-%d")}
        else:
            logging.error('Error Saving file - Encoding object ' + str(obj) + ' failed')
        return return_value


class Context(object):
    """
    Reads (parses) a JSON market data and deals file.
    Also writes out these files once the data has been modified.
    Provides support for parsing grids of dates and working out dynamic dates for the
    given portfolio of deals.
    """

    # class level lookups
    month_lookup = dict((m, i) for i, m in enumerate(calendar.month_abbr) if m)
    offset_lookup = {'M': 'months', 'D': 'days', 'Y': 'years', 'W': 'weeks'}
    reverse_offset = {'months': 'M', 'days': 'D', 'years': 'Y', 'weeks': 'W'}

    def __init__(self, base_currency='USD'):
        self.deals = {'Deals': {'Children': []}}
        self.calibrations = {'CalibrationConfig': {'MarketDataArchiveFile': {}, 'Calibrations': []}}
        self.calendars = ElementTree(Element('Calendars'))
        self.holidays = {}
        self.archive = None
        self.archive_columns = {}

        # the default state of the system
        self.version = None
        self.params = {
            'System Parameters':
                {'Base_Currency': base_currency,
                 'Base_Date': '',
                 'Correlations_Healing_Method': 'Eigenvalue_Raising'
                 },
            'Model Configuration': ModelParams(),
            'Price Factors': {},
            'Price Models': {},
            'Correlations': {},
            'Market Prices': {},
            'Valuation Configuration': {},
            'Bootstrapper Configuration': {}
        }  # type: Dict[str, Union[ModelParams, Dict[Any, Any]]]

        # make sure that there are no default calibration mappings
        self.calibration_process_map = {}
        self.gridparser = get_grid_grammar()

    def parse_grid(self, run_date, max_date, grid, past_max_date=False):
        """
        Construct a set of dates (NOT adjusted to the next business day) as specified in the grid.
        Dates are capped at max_date (but may include the next date after if past_max_date is True)
        """

        parsed_grid = self.gridparser.parseString(grid)[0]
        offsets = parsed_grid.data if isinstance(parsed_grid, utils.Offsets) else [[parsed_grid]]
        fixed_dates = [(run_date + code[0], code[1] if len(code) > 1 else None) for code in offsets] + \
                      [(Timestamp.max, None)]
        dates = set()
        finish = False

        for date_rule, next_date_rule in zip(fixed_dates[:-1], fixed_dates[1:]):
            next_date = date_rule[0]
            if next_date > max_date:
                break
            else:
                dates.add(next_date)
            if date_rule[1]:
                while True:
                    next_date = next_date + date_rule[1]
                    if next_date > max_date:
                        finish = True
                        break
                    if next_date > next_date_rule[0]:
                        break
                    dates.add(next_date)

            if finish:
                break

        if past_max_date:
            dates.add(next_date)

        return dates

    def parse_calendar_file(self, filename):
        """
        Parses the xml calendar file in filename
        """
        self.holidays = {}
        self.calendars = ElementTree(file=filename)
        for elem in self.calendars.iter():
            if elem.attrib.get('Location'):
                holidays = dict(tuple(x.split("|")) for x in elem.attrib['Holidays'].split(', '))
                self.holidays[elem.attrib['Location']] = {
                    'businessday': pd.tseries.offsets.CustomBusinessDay(
                        holidays=holidays.keys(), weekmask=utils.WeekendMap[elem.attrib['Weekends']]),
                    'holidays': holidays}

    def fetch_all_calibration_factors(self, override={}):
        """
        Assumes valid marketdata and calibration config files have been loaded (via parse_json)
        and returns the list of price factors that have mapped price models.
        The return value of this method is suitable to pass to the calibrate_factors method.
        """
        model_factor = {}
        for factor in self.params.get('Price Factors', {}):
            price_factor = utils.check_rate_name(factor)
            model_config = self.params['Model Configuration'].search(utils.Factor(price_factor[0], price_factor[1:]),
                                                                     self.params['Price Factors'][factor])
            model = override.get(model_config, model_config)
            if model:
                subtype = self.params['Price Factors'][factor].get('Sub_Type')
                model_name = utils.Factor(model, price_factor[1:])
                archive_name = utils.Factor(price_factor[0] + (subtype if subtype and subtype != "None" else ''),
                                            price_factor[1:])
                model_factor[factor] = utils.RateInfo(utils.check_tuple_name(model_name),
                                                      utils.check_tuple_name(archive_name),
                                                      self.calibration_process_map.get(model))

        remaining_factor = {}
        remaining_rates = set([col.split(',')[0] for col in self.archive.columns]).difference(model_factor.keys())
        for factor in remaining_rates:
            price_factor = utils.check_rate_name(factor)
            model = self.params['Model Configuration'].search(utils.Factor(price_factor[0], price_factor[1:]), {})
            if model:
                model_name = utils.Factor(model, price_factor[1:])
                archive_name = utils.Factor(price_factor[0], price_factor[1:])
                remaining_factor[factor] = utils.RateInfo(utils.check_tuple_name(model_name),
                                                          utils.check_tuple_name(archive_name),
                                                          self.calibration_process_map.get(model))

        return {'present': model_factor, 'absent': remaining_factor}

    def bootstrap(self):
        """
        Runs all the bootstrappers - this happens in one process with debugging on by default. For multiprocessing
        bootstrapping, call the construct_bootstrapper method directly
        """
        # need to implement ordered dicts in the params obj - TODO
        for bootstrapper_name, params in sorted(self.params['Bootstrapper Configuration'].items()):
            # need parsers here - but for now, can just use the name to know what to do
            try:
                bootstrapper = construct_bootstrapper(bootstrapper_name, params)
            except:
                logging.warning('Cannot execute Bootstrapper for {0} - skipping'.format(bootstrapper_name))
                continue

            # run the bootstrapper on the market prices and store them in the price factors/price models
            bootstrapper.bootstrap(self.params['System Parameters'],
                                   self.params['Price Models'],
                                   self.params['Price Factors'],
                                   self.params['Market Prices'],
                                   self.holidays,
                                   debug=self)

    def calibrate_factors(self, from_date, to_date, factors, smooth=0.0, correlation_cuttoff=0.2,
                          overwrite_correlations=True):
        """
        Assumes a valid calibration JSON configuration file is loaded first, then proceeds to strip out only data
        between from_date and to_date. The calibration rules as specified by the calibration configuration file is then
        applied to the factors given. Note that factors needs to be a list of utils.RateInfo (obtained via a call to
        fetch_all_calibration_factors). Also note that this method overwrites the Price Model section of the config
        file in memory. To save the changes, an explicit call to write_marketdata_json must be made.
        """
        correlation_names = []
        consolidated_df = None
        ak = []
        num_indexes = 0
        num_factors = 0
        total_rates = reduce(operator.concat,
                             [self.archive_columns[rate.archive_name] for rate in factors.values()], [])
        factor_data = utils.filter_data_frame(self.archive, from_date, to_date)[total_rates]

        for rate_name, rate_value in sorted(factors.items()):
            df = factor_data[[col for col in factor_data.columns if col.split(',')[0] == rate_value.archive_name]]
            # now remove spikes
            data_frame = df[np.abs(df - df.median()) <= (smooth * df.std())].interpolate(method='index') \
                if smooth else df
            # log it
            logging.info(
                'Calibrating {0} (archive name {1}) with raw shape {2} and cleaned non-null shape {3}'.format(
                    rate_name, rate_value.archive_name, str(df.shape), str(data_frame.dropna().shape)))

            # calibrate
            try:
                result = rate_value.calibration.calibrate(data_frame, num_business_days=252.0)
            except:
                logging.error('Data errors in factor {0} resulting in flawed calibration - skipping'.format(
                    rate_value.archive_name))
                continue

            # check that it makes sense . . .
            if (np.array(result.correlation).max() > 1) or (np.array(result.correlation).min() < -1) or (
                    result.delta.std() == 0).any():
                logging.error('Data errors in factor {0} resulting in incorrect correlations - skipping'.format(
                    rate_value.archive_name))
                continue

            model_tuple = utils.check_rate_name(rate_value.model_name)
            model_factor = utils.Factor(model_tuple[0], model_tuple[1:])

            # get the correlation name
            process_name, addons = construct_process(model_factor.type, None, result.param).correlation_name

            for sub_factors in addons:
                correlation_names.append(
                    utils.check_tuple_name(utils.Factor(process_name, model_factor.name + sub_factors)))

            consolidated_df = result.delta if consolidated_df is None else pd.concat(
                [consolidated_df, result.delta], axis=1)
            # store the correlation coefficients
            ak.append(result.correlation)

            # store result back in market data file
            self.params['Price Models'][rate_value.model_name] = result.param

            num_indexes += result.delta.shape[1]
            num_factors += rate_value.calibration.num_factors

        a = np.zeros((num_factors, num_indexes))
        rho = consolidated_df.corr()
        offset = 0
        row_num = 0
        for coeff in ak:
            for factor_index, factor in enumerate(coeff):
                a[row_num + factor_index, offset:offset + len(factor)] = factor
            row_num += len(coeff)
            offset += len(factor)

        # cheating here
        factor_correlations = a.dot(rho).dot(a.T).clip(-1.0, 1.0)

        # see if we need to delete the old correlations
        if overwrite_correlations:
            self.params['Correlations'] = {}

        for index1 in range(len(correlation_names) - 1):
            for index2 in range(index1 + 1, len(correlation_names)):
                if np.fabs(factor_correlations[index1, index2]) > correlation_cuttoff:
                    self.params['Correlations'][(correlation_names[index1], correlation_names[index2])] = \
                        factor_correlations[index1, index2]

    def calculate_dependencies(self, options, base_date, base_MTM_dates, calc_dates=True):
        """
        Works out the risk factors (and risk models) in the given set of deals.
        These factors are cross referenced in the marketdata file and matched by
        name. This can be extended as needed.

        Returns the dependant factors, the stochastic models, the full list of
        reset dates and optionally the potential currency settlements
        """

        def get_rates(factor, instrument):
            rates_to_add = {factor: []}
            factor_name = utils.check_tuple_name(factor)

            if factor.type in dependant_fields:
                linked_factors = [utils.Factor(dep_field[1], utils.check_rate_name(
                    self.params['Price Factors'][factor_name][dep_field[0]])) for dep_field in
                                  dependant_fields[factor.type] if
                                  self.params['Price Factors'][factor_name][dep_field[0]]]
                for linked_factor in linked_factors:
                    # add it assuming no dependencies
                    rates_to_add.setdefault(linked_factor, [])
                    # update and dependencies
                    if linked_factor.type in dependant_fields:
                        rates_to_add.update(get_rates(linked_factor, instrument))
                    # link it to the original factor
                    rates_to_add[factor].append(linked_factor)

                    # check that we include any nested factors
                    if linked_factor.type in nested_fields:
                        for i in range(1, len(linked_factor.name) + 1):
                            rates_to_add.update({utils.Factor(linked_factor.type, linked_factor.name[:i]): [
                                utils.Factor(linked_factor.type, linked_factor.name[:i - 1])] if i > 1 else []})

            if factor.type in nested_fields:
                for i in range(1, len(factor.name) + 1):
                    rates_to_add.update({utils.Factor(factor.type, factor.name[:i]): [
                        utils.Factor(factor.type, factor.name[:i - 1])] if i > 1 else []})

            if factor.type in conditional_fields:
                for conditional_factor in conditional_fields[factor.type](instrument,
                                                                          self.params['Price Factors'][factor_name],
                                                                          self.params['Price Factors']):
                    rates_to_add[factor].append(conditional_factor)
                    rates_to_add.update({conditional_factor: []})

            return rates_to_add

        def walk_groups(deals, price_factors, factor_tenors):

            def get_price_factors(rates_to_add, rate_tenors, instrument):
                reval_dates = instrument.get_reval_dates()
                for field_name, factor_types in instrument.factor_fields.items():
                    for field_value in utils.get_fieldname(field_name, instrument.field):
                        for factor in [utils.Factor(factor_type, utils.check_rate_name(field_value))
                                       for factor_type in factor_types]:
                            # store the max tenor for just the factor alone
                            if reval_dates:
                                rate_tenors.setdefault(factor, set()).update({max(reval_dates)})
                            # now look at dependent factors
                            if factor not in rates_to_add:
                                try:
                                    rates_to_add.update(get_rates(factor, instrument))
                                except KeyError as e:
                                    logging.warning('Price Factor {0} not found in market data'.format(e))
                                    if factor.type == 'DiscountRate':
                                        logging.info('Creating default Discount {0}'.format(factor))
                                        self.params['Price Factors'][utils.check_tuple_name(factor)] = OrderedDict(
                                            [('Interest_Rate', '.'.join(factor.name))])
                                        # try again
                                        rates_to_add.update(get_rates(factor, instrument))
                                    else:
                                        logging.error('Skipping Price Factor')

            resets = {base_date}
            children = []
            settlement_currencies = {}

            for node in deals:
                # get the instrument
                instrument = node['instrument']

                if node.get('Ignore') == 'True':
                    continue

                # get a list of children ready to pass back to the parent
                children.append(instrument)

                if node.get('Children'):
                    node_children, node_resets, node_settlements = walk_groups(
                        node['Children'], price_factors, factor_tenors)

                    # sort out dates and calendars
                    instrument.reset(self.holidays)

                    if calc_dates:
                        instrument.finalize_dates(self.parse_grid, base_date, base_MTM_dates, node_children,
                                                  node_resets, node_settlements)
                    # get it's price factors
                    get_price_factors(price_factors, factor_tenors, instrument)

                    # merge dates
                    resets.update(node_resets)
                    for key, value in node_settlements.items():
                        settlement_currencies.setdefault(key, set()).update(value)
                else:
                    # sort out dates and calendars
                    instrument.reset(self.holidays)

                    if calc_dates:
                        instrument.finalize_dates(self.parse_grid, base_date, base_MTM_dates, None, resets,
                                                  settlement_currencies)
                    # get it's price factors
                    get_price_factors(price_factors, factor_tenors, instrument)

            return children, resets, settlement_currencies

        def add_interest_rate(curve_name):
            interest_rate_factor = utils.Factor(
                'InterestRate', utils.check_rate_name(curve_name))

            dependent_factors.update(get_rates(interest_rate_factor, {}))
            for sub_factor in utils.traverse_dependents(interest_rate_factor, dependent_factors):
                dependent_factor_tenors[sub_factor] = reset_dates

        # derived fields are fields that embed other risk factors
        dependant_fields = {'FxRate': [('Interest_Rate', 'InterestRate')],
                            'DiscountRate': [('Interest_Rate', 'InterestRate')],
                            'ForwardPrice': [('Currency', 'FxRate')],
                            'ReferencePrice': [('ForwardPrice', 'ForwardPrice')],
                            'ReferenceVol': [('ForwardPriceVol', 'ForwardPriceVol'),
                                             ('ReferencePrice', 'ReferencePrice')],
                            'InflationRate': [('Price_Index', 'PriceIndex')],
                            'EquityPrice': [('Interest_Rate', 'InterestRate'), ('Currency', 'FxRate')]}

        # nested fields need to include all their children
        nested_fields = {'InterestRate'}

        # conditional fields need to potentially include correlation and fx vol surfaces (e.g. reference prices)
        conditional_fields = {
            'ReferenceVol': lambda instrument, factor_fields, params:
            [utils.Factor('Correlation', tuple('FxRate.{0}.{1}/ReferencePrice.{2}.{0}'.format(
                params[utils.check_tuple_name(utils.Factor('ForwardPrice', (instrument.field['Reference_Type'],)))]
                ['Currency'], instrument.field['Currency'], instrument.field['Reference_Type']).split('.')))] if
            instrument.field['Currency'] != params[utils.check_tuple_name(
                utils.Factor('ForwardPrice', (instrument.field['Reference_Type'],)))]['Currency'] else [],
            'ForwardPrice': lambda instrument, factor_fields, params: [utils.Factor('FXVol', tuple(
                sorted([instrument.field['Currency'], factor_fields['Currency']])))] if
            instrument.field['Currency'] != factor_fields['Currency'] else []}

        # the list of returned factors
        dependent_factors = set()
        stochastic_factors = OrderedDict()
        additional_factors = OrderedDict()

        # complete list of reset dates referenced
        reset_dates = set()
        # complete list of currency settlement dates
        currency_settlement_dates = {}

        # only if we have a portfolio of trades can we calculate its dependencies
        if self.deals:
            # get the reporting currency
            report_currency = options['Currency']
            # get the base currency factor
            base_factor = utils.Factor(
                'FxRate', utils.check_rate_name(self.params['System Parameters']['Base_Currency']))

            # add the base Fx rate
            dependent_factors = get_rates(base_factor, {})

            # store the max date the factor is needed
            dependent_factor_tenors = {}

            # grab all the factor fields in the portfolio
            dependant_deals, reset_dates, currency_settlement_dates = walk_groups(
                self.deals['Deals']['Children'], dependent_factors, dependent_factor_tenors)

            # additional factors from the options passed in
            report_factor = utils.Factor('FxRate', utils.check_rate_name(report_currency))
            report_currency_dependencies = get_rates(report_factor, {})

            # add the base dependencies to the reporting currency
            if report_factor != base_factor:
                report_currency_dependencies[report_factor].append(base_factor)

            dependent_factors.update(report_currency_dependencies)
            # make sure the reporting currency is around till the end
            dependent_factor_tenors[report_factor] = reset_dates
            for reporting_factor in utils.traverse_dependents(report_factor, dependent_factors):
                dependent_factor_tenors[reporting_factor] = reset_dates

            # check if we need to fetch survival data for CVA
            if options.get('CVA'):
                dependent_factors.update(get_rates(
                    utils.Factor('SurvivalProb', utils.check_rate_name(options['CVA']['Counterparty'])),
                    {})
                )
            # check if we need to fetch curve data for FVA
            if options.get('FVA'):
                # add curves
                add_interest_rate(options['FVA']['Funding_Interest_Curve'])
                add_interest_rate(options['FVA']['Risk_Free_Curve'])

                # need to weight the FVA by the survival prob of the counterparty (if defined)
                if 'Counterparty' in options['FVA']:
                    dependent_factors.update(get_rates(
                        utils.Factor('SurvivalProb', utils.check_rate_name(options['FVA']['Counterparty'])), {})
                    )
            # Check deflation
            if options.get('Deflation_Interest_Rate'):
                add_interest_rate(options['Deflation_Interest_Rate'])

            # update the linked factor max tenors
            missing_tenors = {}
            for k, v in dependent_factor_tenors.items():
                for linked_factor in utils.traverse_dependents(k, dependent_factors):
                    missing_tenors.setdefault(linked_factor, set()).update(v)

            # make sure the base currency is always first
            for k, v in dependent_factors.items():
                if k.type == 'FxRate' and k != base_factor and base_factor not in v:
                    v.append(base_factor)

            # now sort the factors taking any factor dependencies into account
            sorted_factors = utils.topological_sort(dependent_factors)
            # merge missing tenors
            for k, v in missing_tenors.items():
                dependent_factor_tenors.setdefault(k, set()).update(v)
            # now get the last tenor for each factor
            dependent_factors = {k: max(dependent_factor_tenors.get(k, reset_dates)) for k in sorted_factors}

            # now lookup the processes
            for factor in sorted_factors:
                stoch_proc = self.params['Model Configuration'].search(factor, self.params['Price Factors'].get(
                    utils.check_tuple_name(factor), {}))
                # might need implied parameters
                additional_factor = self.params['Model Configuration'].additional_factors(stoch_proc, factor)
                if stoch_proc and factor.name[0] != self.params['System Parameters']['Base_Currency']:
                    factor_model = utils.Factor(stoch_proc, factor.name)

                    if utils.check_tuple_name(factor_model) in self.params['Price Models']:
                        stochastic_factors.setdefault(factor_model, factor)
                        if additional_factor:
                            additional_factors.setdefault(factor_model, additional_factor)
                    else:
                        logging.error(
                            'Risk Factor {0} using stochastic process {1} missing in Price Models section'.format(
                                utils.check_tuple_name(factor), stoch_proc))

        return dependent_factors, stochastic_factors, additional_factors, reset_dates, currency_settlement_dates

    def parse_json(self, filename):

        def as_internal(dct):
            if '.Curve' in dct:
                return utils.Curve(dct['.Curve']['meta'], dct['.Curve']['data'])
            elif '.Percent' in dct:
                return utils.Percent(dct['.Percent'])
            elif '.Deal' in dct:
                return construct_instrument(dct['.Deal'], self.params['Valuation Configuration'])
            elif '.Basis' in dct:
                return utils.Basis(dct['.Basis'])
            elif '.Descriptor' in dct:
                return utils.Descriptor(dct['.Descriptor'])
            elif '.DateList' in dct:
                return utils.DateList(OrderedDict([(Timestamp(date), val) for date, val in dct['.DateList']]))
            elif '.DateEqualList' in dct:
                return utils.DateEqualList([[Timestamp(values[0])] + values[1:] for values in dct['.DateEqualList']])
            elif '.CreditSupportList' in dct:
                return utils.CreditSupportList(dct['.CreditSupportList'])
            elif '.DateOffset' in dct:
                return DateOffset(**dct['.DateOffset'])
            elif '.Offsets' in dct:
                return utils.Offsets(dct['.Offsets'])
            elif '.Timestamp' in dct:
                return Timestamp(dct['.Timestamp'])
            elif '.ModelParams' in dct:
                return ModelParams((dct['.ModelParams']['modeldefaults'], dct['.ModelParams']['modelfilters']))
            return dct

        with open(filename, 'rt') as f:
            data = json.load(f, object_hook=as_internal)

        if 'MarketData' in data:
            market_data = data['MarketData']
            correlations = {}
            for rate1, rate_list in market_data['Correlations'].items():
                for rate2, correlation in rate_list.items():
                    correlations.setdefault((rate1, rate2), correlation)

            # update the correlations
            market_data['Correlations'] = correlations
            self.params = market_data
            self.version = data['Version']

        elif 'Deals' in data:
            self.deals = data

        elif 'CalibrationConfig' in data:
            # store the calibration config
            self.calibrations = data

            # load the archive file - must be a tab separated file
            mkt_data_details = self.calibrations['CalibrationConfig']['MarketDataArchiveFile']
            self.archive = pd.read_csv(mkt_data_details['name'], skiprows=mkt_data_details['skiprows'],
                                       sep='\t', index_col=mkt_data_details['index_column'])
            # load the calibration
            self.calibration_process_map = dict((calibration['PriceModel'], construct_calibration_config(calibration))
                                                for calibration in
                                                self.calibrations['CalibrationConfig']['Calibrations'])
            # store a lookup to all columns
            self.archive_columns = {}

            for col in self.archive.columns:
                self.archive_columns.setdefault(col.split(',')[0], []).append(col)

    def write_marketdata_json(self, json_filename):
        # backup old data
        old_correlations = self.params['Correlations']

        # need to serialize out new data
        correlations = {}
        for correlation, value in old_correlations.items():
            correlations.setdefault(correlation[0], {}).setdefault(correlation[1], value)

        # create new keys for json serialization
        self.params['Correlations'] = correlations

        with open(json_filename, 'wt') as f:
            f.write(json.dumps({'MarketData': self.params,
                                'Version': self.version}, separators=(',', ':'), cls=CustomJsonEncoder))

        # restore state
        self.params['Correlations'] = old_correlations

    def write_tradedata_json(self, json_filename):
        with open(json_filename, 'wt') as f:
            f.write(json.dumps(self.deals, separators=(',', ':'), cls=CustomJsonEncoder))

    def write_calibration_json(self, json_filename):
        with open(json_filename, 'wt') as f:
            f.write(json.dumps(self.calibrations, separators=(',', ':'), cls=CustomJsonEncoder))


if __name__ == '__main__':
    pass
